from ctypes import byref, c_int, c_float

from .. import blendmode, surface, rect, video, render, error, dll, hints
from ..stdinc import Uint8, Uint32

from .color import convert_to_color
from .common import raise_sdl_err
from .compat import deprecated, stringify, byteify, isiterable
from .sprite import SoftwareSprite, TextureSprite
from .surface import _get_target_surface
from .window import Window

__all__ = ["set_texture_scale_quality", "Renderer", "Texture"]


# NOTE: The following 3 utility functions probably belong in a separate module

def is_numeric(x):
    try:
        return float(x) == x
    except TypeError:
        return False


def _sanitize_points(points):
    # If first item is numeric, assume flat list of points
    if isinstance(points, rect.SDL_Point) or isinstance(points, rect.SDL_FPoint):
        points = [points]
    elif is_numeric(points[0]):
        if len(points) % 2 != 0:
            raise ValueError("Flat x/y coordinate lists must have even length")
        chunked = []
        for i in range(0, len(points), 2):
            p = (points[i], points[i+1])
            chunked.append(p)
        points = chunked

    type_err = (
        "Points must be specified as either (x, y) tuples or "
        "SDL_Point objects (Got unsupported format '{0}')"
    )
    pts = []
    if not isiterable(points):
        points = [points]
    for p in points:
        if isinstance(p, rect.SDL_Point) or isinstance(p, rect.SDL_FPoint):
            p = (p.x, p.y)
        elif len(p) != 2:
            raise ValueError(type_err.format(str(p)))
        if dll.version < 2010 and any([int(v) != v for v in p]):
            e = "Floating point rendering requires SDL 2.0.10 or newer"
            raise RuntimeError(e + " (got '{0}')".format(str(p)))
        pts.append((p[0], p[1]))
    return pts


def _sanitize_rects(rects):
    type_err = (
        "Rectangles must be specified as either (x, y, w, h) tuples or "
        "SDL_Rect objects (Got unsupported format '{0}')"
    )
    out = []
    if not isiterable(rects) or not isiterable(rects[0]):
        rects = [rects]
    for r in rects:
        if isinstance(r, rect.SDL_Rect) or isinstance(r, rect.SDL_FRect):
            r = (r.x, r.y, r.w, r.h)
        elif len(r) != 4:
            raise ValueError(type_err.format(str(r)))
        if dll.version < 2010 and any([int(v) != v for v in r]):
            e = "Floating point rendering requires SDL 2.0.10 or newer"
            raise RuntimeError(e + " (got '{0}')".format(str(r)))
        out.append((r[0], r[1], r[2], r[3]))
    return out


def _get_texture_size(texture):
    flags = Uint32()
    access, w, h = (c_int(), c_int(), c_int())
    ret = render.SDL_QueryTexture(
        texture, byref(flags), byref(access), byref(w), byref(h)
    )
    if ret < 0:
        raise_sdl_err("getting texture attributes")
    return (w.value, h.value)



def set_texture_scale_quality(method):
    """Sets the default scaling quailty for :obj:`~sdl2.ext.Texture` objects.

    By default, SDL2 uses low-quality nearest-neighbour scaling for all new
    textures. This method lets you change the default scaling method to one of
    the following options:
    
    ========= =============================================================
    Method    Description
    ========= =============================================================
    Nearest   Nearest-neighbour pixel scaling (no filtering)
    Linear    Linear filtering
    Best      Anisotropic filtering (falls back to linear if not available)
    ========= =============================================================

    This function does not apply retroactively, and will only affect textures
    created after it is called.

    Args:
        method (str): The default scaling method to use for SDL textures. Must
            be one of 'nearest', 'linear', or 'best'.

    """
    method = byteify(method.lower())
    if method not in [b"nearest", b"linear", b"best"]:
        raise ValueError(
            "Texture scaling method must be 'nearest', 'linear', or 'best'."
        )
    hints.SDL_SetHint(hints.SDL_HINT_RENDER_SCALE_QUALITY, method)



class Texture(object):
    """A 2D texture to be used with a :obj:`~sdl2.ext.Renderer`.

    In SDL2, textures are 2D images that have been prepared for fast rendering
    with a given renderer. For example, if an SDL surface is converted into a
    texture with a :obj:`~sdl2.ext.Renderer` using the OpenGL backend, the pixel
    data for the surface will internally be converted into an OpenGL texture.

    Once a surface has been converted into a :obj:`Texture`, the surface can be
    safely deleted if no longer needed.

    Args:
        renderer (:obj:`~sdl2.ext.Renderer`): The renderer associated with the
            texture.
        surface (:obj:`~sdl2.SDL_Surface`): An SDL surface from which the
            texture will be created.
    
    """
    def __init__(self, renderer, surface):
        # Validate and get reference to the parent renderer
        self._renderer_ref = None
        if isinstance(renderer, Renderer):
            self._renderer_ref = renderer._renderer_ref
        elif hasattr(renderer, "contents"):
            if isinstance(renderer.contents, render.SDL_Renderer):
                self._renderer_ref = [renderer]
        if self._renderer_ref is None:
            raise TypeError(
                "'renderer' must be a valid Renderer object or a pointer to "
                "an SDL_Renderer."
            )
        # Convert the passed surface into a texture
        surface = _get_target_surface(surface, "surface")
        self._tx = render.SDL_CreateTextureFromSurface(self._renderer, surface)
        if not self._tx:
            raise_sdl_err("creating the texture")

    def __del__(self):
        if hasattr(self, "_tx"):
            self.destroy()

    @property
    def _renderer(self):
        # NOTE: This is sort of a hack to get around a ctypes problem. Basically,
        # if the parent renderer of a texture is destroyed before the texture
        # itself, destroying or doing anything with that texture will result in a
        # segfault. By wrapping the renderer in a List, we can overwrite it with
        # None when destroyed (within the Renderer class) and detect that in any
        # child textures to avoid any memory-unsafe behaviour.
        return self._renderer_ref[0]

    @property
    def tx(self):
        """:obj:`~sdl2.SDL_Texture`: The underlying base SDL texture object.
        Can be used to perform operations with the texture using the base
        PySDL2 bindings.

        """
        if self._tx is None:
            raise RuntimeError(
                "Cannot use a texture after it has been destroyed."
            )
        elif self._renderer_ref[0] is None:
            raise RuntimeError(
                "Cannot use a texture after its parent renderer has been "
                "destroyed."
            )
        return self._tx

    @property
    def size(self):
        """tuple: The width and height (in pixels) of the texture."""
        return _get_texture_size(self.tx)

    def destroy(self):
        """Deletes the texture and frees its associated memory.
        
        When a texture is no longer needed, it should be destroyed using this
        method to free up memory for new textures. After being destroyed, a
        texture can no longer be used.

        """
        if self._tx and self._renderer_ref[0]:
            render.SDL_DestroyTexture(self._tx)
            self._tx = None

    @property
    def scale_mode(self):
        """str: The current scaling mode to use for rendering the texture. Can
        be 'nearest', 'linear', 'best', or 'unknown'.
        
        See :func:`set_texture_scale_quality` for more information.

        .. note::
           For SDL versions older than 2.0.12, the scaling mode will always be
           'unknown'.

        """
        if dll.version < 2012:
            return "unknown"
        modes = ["nearest", "linear", "best"]
        mode_num = c_int()
        ret = render.SDL_GetTextureScaleMode(self.tx, byref(mode_num))
        if ret < 0:
            raise_sdl_err("retrieving the texture scaling mode")
        try:
            return modes[mode_num.value]
        except IndexError:
            return "unknown"

    def set_scale_mode(self, mode):
        """Sets a custom scaling method to use for rendering the texture.

        This method overrides the default texture scaling method specified by
        :func:`set_texture_scale_quality`, the documentation for which describes
        the different possible scaling modes.

        .. note::
           Support for custom per-texture scaling modes is only available in
           SDL2 2.0.12 and up. As such, this method has no effect when used with
           earlier releases of SDL2.

        Args:
            mode (str): The scaling method to use for the current texture. Must
                be one of 'nearest', 'linear', or 'best'.

        """
        if dll.version < 2012:
            return None
        modes = {
            "nearest": render.SDL_ScaleModeNearest,
            "linear": render.SDL_ScaleModeLinear,
            "best": render.SDL_ScaleModeBest,
        }
        mode = mode.lower()
        if mode not in modes.keys():
            raise ValueError(
                "Texture scaling mode must be either 'nearest', 'linear', "
                "or 'best'."
            )
        ret = render.SDL_SetTextureScaleMode(self.tx, modes[mode])
        if ret < 0:
            raise_sdl_err("setting the texture scaling mode")



class Renderer(object):
    """A rendering context for SDL2 windows and sprites.

    A ``Renderer`` can be created from an SDL window, an SDL Surface, or a
    :obj:`~sdl2.ext.SoftwareSprite`. Depending on the settings and operating
    system, this rendering context can use either hardware or software
    acceleration.

    A useful feature of SDL renderers is that that the logical size of the
    rendering context can be different than the actual size (in pixels) of its
    target window or surface. For example, the renderer for a 1024x768 window
    can be set to have a logical size of 640x480, improving performance at the
    cost of image quality. The rendered content will be automatically scaled to
    fit the target, and will be centered with black bars on either side in the
    case of an aspect ratio mismatch.

    If creating a rendering context from a window, you can customize the
    renderer using flags to request different settings:

    =============================== ===========================================
    Flag                            Description
    =============================== ===========================================
    ``SDL_RENDERER_SOFTWARE``       Requests a software-accelerated renderer.
    ``SDL_RENDERER_ACCELERATED``    Requests a hardware-accelerated renderer.
    ``SDL_RENDERER_PRESENTVSYNC``   Enables vsync support for :meth:`present`.
    ``SDL_RENDERER_TARGETTEXTURE``  Requests support for rendering to texture.
    =============================== ===========================================

    To combine multiple flags, you can use a bitwise OR to combine two or more
    together before passing them to the `flags` argument::

      render_flags = (
          sdl2.SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC
      )
      sdl2.ext.Renderer(window, flags=render_flags)

    By default, SDL2 will choose the first renderer backend that supports all
    the requested flags. However, you can also request a specific rendering
    backend by name (e.g. 'opengl', 'opengles2', 'metal', 'direct3d', etc.),
    giving you more control but likely making your code less cross-platform.

    Args:
        target (:obj:`~sdl2.ext.Window`, :obj:`~sdl2.SDL_Surface`): The target
            window or surface from which to create the rendering context.
        backend (str or int, optional): The name of the specific backend to use
            for the new rendering context (e.g. 'opengl'). Defaults to letting
            SDL2 decide. If ``target`` is not an SDL window, this argument has
            no effect.
        logical_size (tuple, optional): The initial logical size (in pixels) of
            the rendering context as a ``(w, h)`` tuple. Defaults to the size
            of the target window or surface.
        flags (int, optional): The requested features and settings for the
            new rendering context. Defaults to requesting a hardware-accelerated
            context. If ``target`` is not an SDL window, this argument has no
            effect.
 
    Raises:
        RuntimeError: If a requested rendering backend is not available.

    """

    def __init__(self, target, backend=-1, logical_size=None,
                 flags=render.SDL_RENDERER_ACCELERATED):
        self._renderer_ref = None
        self.rendertarget = None

        available = self._get_render_drivers()
        if isinstance(backend, str):
            if backend.lower() in available:
                index = available.index(backend.lower())
            else:
                e1 = "'{0}' is not a supported renderer on this system. "
                e2 = "The renderer backend must be one of the following: "
                raise RuntimeError(e1.format(backend) + e2 + str(available))
        else:
            index = backend

        _size = None
        if isinstance(target, Window):
            _renderer = render.SDL_CreateRenderer(target.window, index, flags)
            _size = target.size
        elif isinstance(target, video.SDL_Window):
            _renderer = render.SDL_CreateRenderer(target, index, flags)
            w, h = c_int(0), c_int(0)
            video.SDL_GetWindowSize(target, byref(w), byref(h))
            _size = (w.value, h.value)
        elif isinstance(target, SoftwareSprite):
            _renderer = render.SDL_CreateSoftwareRenderer(target.surface)
            _size = target.size
        elif isinstance(target, surface.SDL_Surface):
            _renderer = render.SDL_CreateSoftwareRenderer(target)
            _size = (target.w, target.h)
        elif "SDL_Surface" in str(type(target)):
            _renderer = render.SDL_CreateSoftwareRenderer(target.contents)
            _size = (target.contents.w, target.contents.h)
        else:
            raise TypeError("unsupported target type")
        if not _renderer:
            raise_sdl_err("creating the SDL renderer")
        error.SDL_ClearError()  # Clear any errors from renderer selection
        self._renderer_ref = [_renderer]

        self.rendertarget = target
        self.color = (0, 0, 0, 0)  # Set black as the default draw color
        self.logical_size = _size
        self._original_logical_size = self.logical_size
        if logical_size is not None:
            self.logical_size = logical_size

    def __del__(self):
        if self._renderer_ref is not None:
            self.destroy()

    def _get_render_drivers(self):
        renderers = []
        drivers = render.SDL_GetNumRenderDrivers()
        for x in range(drivers):
            info = render.SDL_RendererInfo()
            ret = render.SDL_GetRenderDriverInfo(x, info)
            if ret != 0:
                renderers.append(stringify(info.name))
            error.SDL_ClearError()
        return renderers

    @property
    def sdlrenderer(self):
        """:obj:`~sdl2.SDL_Renderer`: The underlying base SDL renderer object.
        Can be used to perform operations with the renderer using the base
        PySDL2 bindings.

        """
        if self._renderer_ref[0] is None:
            raise RuntimeError(
                "Cannot use a renderer after it has been destroyed."
            )
        return self._renderer_ref[0]

    @property
    @deprecated
    def renderer(self):
        return self.sdlrenderer

    @property
    def logical_size(self):
        """tuple: The logical size of the rendering context (in pixels), as a
        ``(width, height)`` tuple.

        """
        w, h = c_int(0), c_int(0)
        render.SDL_RenderGetLogicalSize(self.sdlrenderer, byref(w), byref(h))
        return w.value, h.value

    @logical_size.setter
    def logical_size(self, size):
        width, height = size
        ret = render.SDL_RenderSetLogicalSize(self.sdlrenderer, width, height)
        if ret != 0:
            raise_sdl_err("setting the logical size of the renderer")

    @property
    def color(self):
        """:obj:`~sdl2.ext.Color`: The current drawing color of the renderer."""
        r, g, b, a = Uint8(0), Uint8(0), Uint8(0), Uint8(0)
        ret = render.SDL_GetRenderDrawColor(self.sdlrenderer, byref(r), byref(g),
                                            byref(b), byref(a))
        if ret < 0:
            raise_sdl_err("retrieving the drawing color of the renderer")
        return convert_to_color((r.value, g.value, b.value, a.value))

    @color.setter
    def color(self, value):
        c = convert_to_color(value)
        ret = render.SDL_SetRenderDrawColor(self.sdlrenderer, c.r, c.g, c.b, c.a)
        if ret < 0:
            raise_sdl_err("setting the drawing color of the renderer")

    @property
    def blendmode(self):
        """int: The blend mode used for :meth:`fill` and :meth:`line` drawing
        operations. This value can be any of the following constants:
        
        ========================= ====================================
        Flag                      Description
        ========================= ====================================
        ``SDL_BLENDMODE_NONE``    No blending
        ``SDL_BLENDMODE_BLEND``   Alpha channel blending
        ``SDL_BLENDMODE_ADD``     Additive blending
        ``SDL_BLENDMODE_MOD``     Color modulation
        ``SDL_BLENDMODE_MUL``     Color multiplication (SDL >= 2.0.12)
        ========================= ====================================

        """
        mode = blendmode.SDL_BlendMode()
        ret = render.SDL_GetRenderDrawBlendMode(self.sdlrenderer, byref(mode))
        if ret < 0:
            raise_sdl_err("retrieving the blend mode for the renderer")
        return mode

    @blendmode.setter
    def blendmode(self, value):
        ret = render.SDL_SetRenderDrawBlendMode(self.sdlrenderer, value)
        if ret < 0:
            raise_sdl_err("setting the blend mode for the renderer")

    @property
    def scale(self):
        """tuple: The x/y scaling factors applied to all drawing coordinates
        before rendering, in the format ``(scale_x, scale_y)``. These can be
        used to facilitate resolution-independent drawing.
        
        """
        sx = c_float(0.0)
        sy = c_float(0.0)
        render.SDL_RenderGetScale(self.sdlrenderer, byref(sx), byref(sy))
        return sx.value, sy.value

    @scale.setter
    def scale(self, value):
        if any([s <= 0 for s in value]):
            raise ValueError("Scaling factors must be greater than zero.")
        sx, sy = value
        ret = render.SDL_RenderSetScale(self.sdlrenderer, sx, sy)
        if ret != 0:
            raise_sdl_err("setting the scaling factors for the renderer")

    def destroy(self):
        """Destroys the renderer and any associated textures.

        When a renderer is no longer needed, it should be destroyed using this
        method to free up its associated memory. After being destroyed, a
        renderer can no longer be used.
        
        """
        if self._renderer_ref[0]:
            render.SDL_DestroyRenderer(self._renderer_ref[0])
            self._renderer_ref[0] = None
            self.rendertarget = None

    def reset_logical_size(self):
        """Resets the logical size of the renderer to its original value."""
        self.logical_size = self._original_logical_size

    def clear(self, color=None):
        """Clears the rendering surface with a given color.

        Args:
            color (:obj:`~sdl2.ext.Color`, optional): The color with which to
                clear the entire rendering context. If not specified, the
                renderer's current :attr:`color` will be used.

        """
        if color is None:
            ret = render.SDL_RenderClear(self.sdlrenderer)
        else:
            tmp = self.color
            self.color = color
            ret = render.SDL_RenderClear(self.sdlrenderer)
            self.color = tmp
        if ret < 0:
           raise_sdl_err("clearing the rendering context")

    def copy(self, src, srcrect=None, dstrect=None, angle=0, center=None,
             flip=render.SDL_FLIP_NONE):
        """Copies (blits) a texture to the rendering context.

        If the source texture is an :obj:`~sdl2.SDL_Surface`, you will need
        to convert it into a :obj:`~sdl2.ext.Texture` first before it can be
        copied to the rendering surface.

        The source texture can be flipped horizontally or vertically when
        being copied to the rendering context using one of the following
        flags:

        ========================= ===================================
        Flag                      Description
        ========================= ===================================
        ``SDL_FLIP_NONE``         Does not flip the source (default)
        ``SDL_FLIP_HORIZONTAL``   Flips the source horizontally
        ``SDL_FLIP_VERTICAL``     Flips the source vertically
        ========================= ===================================

        .. note::
           Subpixel rendering (i.e. using floats as pixel coordinates) requires
           SDL 2.0.10 or newer.

        Args:
            src (:obj:`~sdl2.ext.Texture`, :obj:`~sdl2.SDL_Texture`): The source
                texture to copy to the rendering surface.
            srcrect (tuple, optional): An ``(x, y, w, h)`` rectangle defining
                the subset of the source texture to copy to the rendering
                surface. Defaults to copying the entire source texture.
            dstrect (tuple, optional): An ``(x, y, w, h)`` rectangle
                defining the region of the rendering surface to which the 
                source texture will be copied. Alternatively, if only
                ``(x, y)`` coordinates are provided, the width and height of
                the source rectangle will be used. Defaults to stretching
                the source across the entire rendering context.
            angle (float, optional): The clockwise rotation (in degrees) to
                apply to the destination rectangle. Defaults to no rotation.
            center (tuple, optional): The point around with the destination
                rectangle will be rotated. Defaults to the center of the
                destination rectangle.
            flip (int, optional): A flag indicating whether the source should
                be flipped (horizontally or vertically) when rendering to the
                render context. Defaults to no flipping.

        """
        if dll.version < 2010:
            render_copy = render.SDL_RenderCopyEx
            Point = rect.SDL_Point
            Rect = rect.SDL_Rect
        else:
            render_copy = render.SDL_RenderCopyExF
            Point = rect.SDL_FPoint
            Rect = rect.SDL_FRect

        if isinstance(src, TextureSprite):
            texture = src.texture
            angle = angle if angle != 0 else src.angle
            center = center if center else src.center
            flip = flip if flip != 0 else src.flip
        elif isinstance(src, Texture):
            texture = src.tx
        elif isinstance(src, render.SDL_Texture):
            texture = src
        else:
            raise TypeError("src must be a Texture object or an SDL_Texture")

        if srcrect:
            x, y, w, h = _sanitize_rects([srcrect])[0]
            srcrect = rect.SDL_Rect(int(x), int(y), int(w), int(h))

        if dstrect:
            if len(dstrect) == 2:
                x, y = _sanitize_points([dstrect])[0]
                if srcrect:
                    w, h = (srcrect.w, srcrect.h)
                else:
                    w, h = _get_texture_size(texture)
            elif len(dstrect) == 4:
                x, y, w, h = _sanitize_rects([dstrect])[0]
            dstrect = Rect(x, y, w, h)

        if center:
            x, y = _sanitize_points(center)[0]
            center = Point(x, y)

        ret = render_copy(
            self.sdlrenderer, texture, srcrect, dstrect, angle, center, flip
        )
        if ret < 0:
            raise_sdl_err("copying the texture to the rendering context")

    def blit(self, src, srcrect=None, dstrect=None, angle=0, center=None,
             flip=render.SDL_FLIP_NONE):
        """Copies a texture to the rendering context.
        
        An alias for the :meth:`copy` method.

        """
        self.copy(src, srcrect, dstrect, angle, center, flip)

    def present(self):
        """Presents the current rendering surface to the screen.

        Because SDL renderers use batch rendering (i.e. they have a separate
        backbuffer that is drawn to and buffer shown on screen which are
        switched when this function is called), any drawing operations
        performed with the renderer will not take effect until this method
        is called.

        It is recommended that you clear and redraw the contents of the
        rendering context every time before this method is called, as the
        contents of the buffers are not guaranteed to remain the same between
        repeat presentations.
        
        """
        render.SDL_RenderPresent(self.sdlrenderer)

    def draw_line(self, points, color=None):
        """Draws one or more connected lines on the rendering context.

        .. note::
           Subpixel rendering (i.e. using floats as pixel coordinates) requires
           SDL 2.0.10 or newer.

        Args:
            points (list): A list of 2 or more ``(x, y)`` coordinates or
                :obj:`~sdl2.SDL_Point` objects defining the set of connected
                lines to draw.
            color (:obj:`~sdl2.ext.Color`, optional): The color with which to
                draw the lines. If not specified, the renderer's current
                :attr:`color` will be used.

        """
        if dll.version < 2010:
            draw_lines = render.SDL_RenderDrawLines
            Point = rect.SDL_Point
        else:
            draw_lines = render.SDL_RenderDrawLinesF
            Point = rect.SDL_FPoint

        points = _sanitize_points(points)
        if len(points) < 2:
            raise ValueError("At least two (x, y) points are required.")
        sdlpts = []
        for p in points:
            x, y = p
            sdlpts.append(Point(x, y))
        points_ptr = (Point * len(points))(*sdlpts)

        if color is None:
            ret = draw_lines(self.sdlrenderer, points_ptr, len(points))
        else:
            tmp = self.color
            self.color = color
            ret = draw_lines(self.sdlrenderer, points_ptr, len(points))
            self.color = tmp

        if ret < 0:
            raise_sdl_err("drawing lines to the renderer")

    def draw_point(self, points, color=None):
        """Draws one or more points on the rendering context.

        .. note::
           Subpixel rendering (i.e. using floats as pixel coordinates) requires
           SDL 2.0.10 or newer.

        Args:
            points (list): A list of ``(x, y)`` coordinates or
                :obj:`~sdl2.SDL_Point` objects defining the set of points to
                draw.
            color (:obj:`~sdl2.ext.Color`, optional): The color with which to
                draw the points. If not specified, the renderer's current
                :attr:`color` will be used.

        """
        if dll.version < 2010:
            draw_points = render.SDL_RenderDrawPoints
            Point = rect.SDL_Point
        else:
            draw_points = render.SDL_RenderDrawPointsF
            Point = rect.SDL_FPoint

        points = _sanitize_points(points)
        sdlpts = []
        for p in points:
            x, y = p
            sdlpts.append(Point(x, y))
        points_ptr = (Point * len(points))(*sdlpts)

        if color is None:
            ret = draw_points(self.sdlrenderer, points_ptr, len(points))
        else:
            tmp = self.color
            self.color = color
            ret = draw_points(self.sdlrenderer, points_ptr, len(points))
            self.color = tmp

        if ret < 0:
            raise_sdl_err("drawing points to the renderer")

    def draw_rect(self, rects, color=None):
        """Draws one or more rectangles on the rendering context.

        Rectangles can be specified as 4-item ``(x, y, w, h)`` tuples,
        :obj:`~sdl2.rect.SDL_Rect` objects, or a list containing multiple
        rectangles in either format.

        .. note::
           Subpixel rendering (i.e. using floats as pixel coordinates) requires
           SDL 2.0.10 or newer.

        Args:
            rects (tuple, :obj:`~sdl2.SDL_Rect`, list): The rectangle(s) to draw
                to the rendering context.
            color (:obj:`~sdl2.ext.Color`, optional): The color with which to
                draw the rectangle(s). If not specified, the renderer's current
                :attr:`color` will be used.

        """
        if dll.version < 2010:
            draw_rects = render.SDL_RenderDrawRects
            Rect = rect.SDL_Rect
        else:
            draw_rects = render.SDL_RenderDrawRectsF
            Rect = rect.SDL_FRect

        rects = _sanitize_rects(rects)
        sdlrects = []
        for r in rects:
            x, y, w, h = r
            sdlrects.append(Rect(x, y, w, h))
        rects_ptr = (Rect * len(rects))(*sdlrects)

        if color is None:
            ret = draw_rects(self.sdlrenderer, rects_ptr, len(rects))
        else:
            tmp = self.color
            self.color = color
            ret = draw_rects(self.sdlrenderer, rects_ptr, len(rects))
            self.color = tmp
    
        if ret < 0:
            raise_sdl_err("drawing rectangles to the renderer")

    def fill(self, rects, color=None):
        """Fills one or more rectangular regions the rendering context.

        Fill regions can be specified as 4-item ``(x, y, w, h)`` tuples,
        :obj:`~sdl2.rect.SDL_Rect` objects, or a list containing multiple
        rectangles in either format.

        .. note::
           Subpixel rendering (i.e. using floats as pixel coordinates) requires
           SDL 2.0.10 or newer.

        Args:
            rects (tuple, :obj:`~sdl2.SDL_Rect`, list): The rectangle(s) to fill
                within the rendering context.
            color (:obj:`~sdl2.ext.Color`, optional): The color with which to
                fill the rectangle(s). If not specified, the renderer's current
                :attr:`color` will be used.

        """
        if dll.version < 2010:
            fill_rects = render.SDL_RenderFillRects
            Rect = rect.SDL_Rect
        else:
            fill_rects = render.SDL_RenderFillRectsF
            Rect = rect.SDL_FRect

        rects = _sanitize_rects(rects)
        sdlrects = []
        for r in rects:
            x, y, w, h = r
            sdlrects.append(Rect(x, y, w, h))
        rects_ptr = (Rect * len(rects))(*sdlrects)

        if color is None:
            ret = fill_rects(self.sdlrenderer, rects_ptr, len(rects))
        else:
            tmp = self.color
            self.color = color
            ret = fill_rects(self.sdlrenderer, rects_ptr, len(rects))
            self.color = tmp
    
        if ret < 0:
            raise_sdl_err("filling rectangles in the renderer")



import os
import sys
from ctypes import c_int, byref
import pytest
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_VIDEO
from sdl2 import surface, render

sdlgfx = pytest.importorskip("sdl2.sdlgfx")



@pytest.fixture(scope="module")
def with_sdl():
    if SDL_Init(SDL_INIT_VIDEO) != 0:
        raise RuntimeError("Video subsystem not supported")
    yield
    SDL_Quit()

@pytest.fixture
def test_surf(with_sdl):
    w, h = 480, 320
    sf = surface.SDL_CreateRGBSurface(0, w, h, 32, 0, 0, 0, 0)
    assert isinstance(sf.contents, surface.SDL_Surface)
    yield sf
    surface.SDL_FreeSurface(sf)

@pytest.fixture
def software_renderer(with_sdl):
    height, width = (480, 320)
    sdl2.SDL_ClearError()
    # Create a new SDL surface and make it the target of a new renderer
    target = surface.SDL_CreateRGBSurface(0, height, width, 32, 0, 0, 0, 0)
    assert sdl2.SDL_GetError() == b""
    renderer = render.SDL_CreateSoftwareRenderer(target)
    assert sdl2.SDL_GetError() == b""
    # Return the renderer and surface for the test
    yield (renderer, target)
    # Free memory for the renderer and surface once we're done
    sdl2.SDL_DestroyRenderer(renderer)
    sdl2.SDL_FreeSurface(target)

RED_U32 = 0xFF0000FF
WHITE_U32 = 0xFFFFFFFF


class TestFramerate(object):
    __tags__ = ["sdl", "sdlgfx"]

    def test_SDL_initFramerate(self):
        manager = sdlgfx.FPSManager()
        sdlgfx.SDL_initFramerate(manager)
        assert manager.framecount == 0
        assert manager.rate == sdlgfx.FPS_DEFAULT

    def test_SDL_getsetFramerate(self):
        manager = sdlgfx.FPSManager()
        sdlgfx.SDL_initFramerate(manager)
        ret = sdlgfx.SDL_setFramerate(manager, 24)
        fps = sdlgfx.SDL_getFramerate(manager)
        assert ret == 0
        assert manager.rate == 24
        assert fps == 24

    def test_SDL_getFramecount(self):
        manager = sdlgfx.FPSManager()
        sdlgfx.SDL_initFramerate(manager)
        assert sdlgfx.SDL_getFramecount(manager) == 0
        sdlgfx.SDL_framerateDelay(manager)
        assert sdlgfx.SDL_getFramecount(manager) == 1

    @pytest.mark.xfail(reason="Timing can be super wonky on CI platforms")
    def test_SDL_framerateDelay(self):
        manager = sdlgfx.FPSManager()
        sdlgfx.SDL_initFramerate(manager)
        ret = sdlgfx.SDL_setFramerate(manager, 100)
        assert ret == 0
        sdlgfx.SDL_framerateDelay(manager)
        assert manager.framecount == 1
        frametimes = []
        for i in range(10):
            msec_since_last = sdlgfx.SDL_framerateDelay(manager)
            frametimes.append(msec_since_last)
        avg_framerate = round(sum(frametimes) / 10)
        assert abs(avg_framerate - 10) < 2
        assert manager.framecount == 11


class TestDrawing(object):
    __tags__ = ["sdl", "sdlgfx"]

    def test_pixel(self, software_renderer):
        renderer, sf = software_renderer
        ret = sdlgfx.pixelColor(renderer, 5, 5, RED_U32)
        assert ret == 0
        ret = sdlgfx.pixelRGBA(renderer, 5, 5, 255, 255, 255, 255)
        assert ret == 0

    @pytest.mark.skip("not implemented")
    def test_hline(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_vline(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_rectangle(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_roundedRectangle(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_box(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_roundedBox(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_line(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_aaline(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_thickLine(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_circle(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_arc(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_aacircle(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_filledCircle(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_ellipse(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_aaellipse(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_filledEllipse(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_pie(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_filledPie(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_trigon(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_aatrigon(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_filledTrigon(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_polygon(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_aapolygon(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_filledPolygon(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_texturedPolygon(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_bezier(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_gfxPrimitivesSetFont(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_gfxPrimitivesSetFontRotation(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_character(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_string(self):
        pass


class TestRotoZoom(object):
    __tags__ = ["sdl", "sdlgfx"]

    @pytest.mark.skip("not implemented")
    def test_rotozoomSurface(self, test_surf):
        pass

    @pytest.mark.skip("not implemented")
    def test_rotozoomSurfaceXY(self, test_surf):
        pass

    @pytest.mark.skip("not implemented")
    def test_rotozoomSurfaceSize(self, test_surf):
        pass

    @pytest.mark.skip("not implemented")
    def test_rotozoomSurfaceSizeXY(self, test_surf):
        pass

    def test_zoomSurface(self, test_surf):
        zoom_sf = sdlgfx.zoomSurface(test_surf, 1.5, 2.0, 0)
        assert isinstance(zoom_sf.contents, surface.SDL_Surface)
        assert zoom_sf.contents.w == test_surf.contents.w * 1.5
        assert zoom_sf.contents.h == test_surf.contents.h * 2.0
        surface.SDL_FreeSurface(zoom_sf)

    def test_zoomSurfaceSize(self, test_surf):
        out_w, out_h = (c_int(0), c_int(0))
        w, h = (test_surf.contents.w, test_surf.contents.h)
        sdlgfx.zoomSurfaceSize(w, h, 1.5, 2.0, byref(out_w), byref(out_h))
        assert out_w.value == w * 1.5
        assert out_h.value == h * 2.0

    def test_shrinkSurface(self, test_surf):
        shrunk_sf = sdlgfx.shrinkSurface(test_surf, 3, 2)
        assert isinstance(shrunk_sf.contents, surface.SDL_Surface)
        assert shrunk_sf.contents.w == test_surf.contents.w / 3
        assert shrunk_sf.contents.h == test_surf.contents.h / 2
        surface.SDL_FreeSurface(shrunk_sf)

    def test_rotateSurface90Degrees(self, test_surf):
        rotsf = sdlgfx.rotateSurface90Degrees(test_surf, 1)
        assert isinstance(rotsf.contents, surface.SDL_Surface)
        assert rotsf.contents.w == test_surf.contents.h
        assert rotsf.contents.h == test_surf.contents.w
        surface.SDL_FreeSurface(rotsf)



import sys
import pytest
from sdl2 import SDL_Window, SDL_ClearError
from sdl2 import ext as sdl2ext
from sdl2 import messagebox as mb


def test_MessageBoxTheme(with_sdl):
    # Test using RGB color tuples
    theme = sdl2ext.MessageBoxTheme(text=(255, 255, 255))
    text_col = theme._get_theme().colors[1]
    assert text_col.r == 255 and text_col.g == 255

    # Test using Color objects
    BLACK = sdl2ext.Color(0, 0, 0)
    theme = sdl2ext.MessageBoxTheme(btn=BLACK)
    btn_col = theme._get_theme().colors[3]
    assert btn_col.r == 0 and btn_col.b == 0

    # Test exceptions on bad input
    with pytest.raises(TypeError):
        sdl2ext.MessageBoxTheme(bg=(255, 255))
    with pytest.raises(ValueError):
        sdl2ext.MessageBoxTheme(bg=(256, 255, 255))

def test_MessageBox(with_sdl):
    # Test initialization of the class
    box = sdl2ext.MessageBox(
        "Test", "Did it work?", ["Yes", "No"], default="Yes", msgtype="info"
    )
    sdl_box = box._get_msgbox()
    assert sdl_box.flags & mb.SDL_MESSAGEBOX_INFORMATION
    assert sdl_box.title.decode('utf-8') == "Test"
    assert sdl_box.message.decode('utf-8') == "Did it work?"
    assert sdl_box.numbuttons == 2
    b1 = sdl_box.buttons.contents  # Only gets first button for some reason
    assert b1.text.decode('utf-8') == "Yes"
    assert b1.flags & mb.SDL_MESSAGEBOX_BUTTON_RETURNKEY_DEFAULT

    # Test internal generation w/ an associated window
    win = sdl2ext.Window(b"Test", (640, 480))
    sdl_box = box._get_msgbox(win)
    assert isinstance(sdl_box.window.contents, SDL_Window)
    win.close()

    # Test initialization with theme
    mbtheme = sdl2ext.MessageBoxTheme(text=(255, 0, 0))
    box2 = sdl2ext.MessageBox(
        "Test", "Did it work?", ["Yes", "No"], theme=mbtheme
    )
    sdl_box2 = box2._get_msgbox()
    text_col = sdl_box2.colorScheme.contents.colors[1]
    assert text_col.r == 255 and text_col.g == 0

    # Test exceptions on bad input
    with pytest.raises(TypeError):
        sdl2ext.MessageBox("Title", "Some text", "A button")
    with pytest.raises(ValueError):
        sdl2ext.MessageBox("Title", "Some text", [])
    with pytest.raises(ValueError):
        box._get_window_pointer("not a window")

@pytest.mark.skip("not implemented, requires GUI interaction")
def test_show_messagebox(self):
    # Could implement a test using mock, but not sure how useful that'd be
    pass

@pytest.mark.skip("not implemented, requires GUI interaction")
def test_show_alert(self):
    pass



import sys
import pytest
from sdl2 import platform


def test_SDL_GetPlatform():
    retval = platform.SDL_GetPlatform()
    if sys.platform in ("win32", "cygwin"):
        assert retval == b"Windows"
    elif sys.platform.startswith("linux"):
        assert retval == b"Linux"
    elif sys.platform.startswith("freebsd"):
        assert retval == b"FreeBSD"
    elif sys.platform.startswith("darwin"):
        assert retval == b"Mac OS X"
    # Do not check others atm, since we are unsure about what Python will
    # return here.



import sys
import os
import pytest
from ctypes import c_int, byref
from sdl2 import power


def test_SDL_GetPowerInfo():
    has_battery = [
        power.SDL_POWERSTATE_ON_BATTERY,
        power.SDL_POWERSTATE_CHARGING,
        power.SDL_POWERSTATE_CHARGED
    ]
    remaining, pct = c_int(), c_int()
    state = power.SDL_GetPowerInfo(byref(remaining), byref(pct))
    if state in has_battery:
        assert pct.value <= 100
        assert pct.value > 0
    else:
        assert remaining.value == -1
        assert pct.value == -1



import ctypes

__all__ = ["CTypesView", "to_ctypes", "to_list", "to_tuple", "create_array",
           "MemoryView"]


# Hack around an import error using relative import paths in Python 2.7
_ARRAY = __import__("array")


def to_tuple(dataseq):
    """Converts a ``ctypes`` array to a tuple."""
    return tuple(dataseq)


def to_list(dataseq):
    """Converts a ``ctypes`` array to a list."""
    return list(dataseq)


def to_ctypes(dataseq, dtype, mcount=0):
    """Converts an arbitrary sequence to a ``ctypes`` array of a given type.

    Args:
        dataseq: A sequence to convert to a ``ctypes`` array.
        dtype (class): The ``ctypes`` data type to use for the array (e.g.
            ``ctypes.c_uint8``).
        mcount (int, optional): The number of elements in ``dataseq``. If
            not specified, this will be inferred automatically.

    Returns:
        tuple: A tuple in the form ``(valset, count)``, where ``valset`` is
        the converted ``ctypes`` array and ``count`` is the number of
        elements in the array.

    Raises:
        TypeError: if any elements in the passed sequence do not
            match the specified type.

    """
    if mcount > 0:
        count = mcount
    else:
        count = len(dataseq)
    if isinstance(dataseq, CTypesView):
        itemsize = ctypes.sizeof(dtype)
        if itemsize == 1:
            dataseq = dataseq.to_bytes()
        elif itemsize == 2:
            dataseq = dataseq.to_uint16()
        elif itemsize == 4:
            dataseq = dataseq.to_uint32()
        elif itemsize == 8:
            dataseq = dataseq.to_uint64()
        else:
            raise TypeError("unsupported data type for the passed CTypesView")
    valset = (count * dtype)(*dataseq)
    return valset, count


def create_array(obj, itemsize):
    """Creates an :obj:`array.array` copy of a given object.

    Args:
        obj: The object from which the array will be created.
        itemsize: The size (in bytes) of each item in the given object.

    Returns:
        :obj:`array.array`: The array created from the object.

    """
    if itemsize == 1:
        return _ARRAY.array("B", obj)
    elif itemsize == 2:
        return _ARRAY.array("H", obj)
    elif itemsize == 4:
        return _ARRAY.array("I", obj)
    elif itemsize == 8:
        return _ARRAY.array("d", obj)
    else:
        raise TypeError("unsupported data type")


class CTypesView(object):
    """A proxy for accessing byte-wise ``ctypes`` data types.

    This class provides read-write access for arbitrary ``ctypes`` objects
    that are iterable. In case the object does not provide a :func:`buffer`
    interface for direct access, a ``CTypesView`` can copy the object's contents
    into an internal buffer from which data can be retrieved.

    Depending on the item type stored in the iterable object, you might
    need to manually specify the object's item size (in bytes). Additionally,
    you may need to manually specify the number of items in the iterable if
    it does not properly return its length using the ``len`` function.

    For certain types, such as the bytearray, the original object must not be
    reassigned after being encapsuled and used in ctypes bindings if the
    contents are not copied.

    Args:
        obj: An arbitrary iterable ``ctypes`` object to access.
        itemsize (int, optional): The size (in bytes) of each item of the
            iterable object. Defaults to ``1``.
        docopy (bool, optional): If True, the view will be created for a copy
            of the object and will not modify the original ``ctypes`` instance.
            Defalts to False (original object will be modified).
        objsize (int, optional): The number of items in the iterable object.
            If not specified, the number of items will try to be inferred
            automatically. Defaults to ``None`` (automatic inferrence).

    """
    def __init__(self, obj, itemsize=1, docopy=False, objsize=None):
        self._obj = obj
        self._isshared = True
        self._view = None
        self._itemsize = itemsize
        self._create_view(itemsize, bool(docopy), objsize)

    def _create_view(self, itemsize, docopy, objsize):
        """Creates the view on the specified object."""
        self._isshared = not docopy
        bsize = 0
        if objsize is None:
            bsize = len(self._obj) * itemsize
        else:
            bsize = objsize * itemsize

        if docopy:
            self._obj = create_array(self._obj, itemsize)
        try:
            self._view = (ctypes.c_ubyte * bsize).from_buffer(self._obj)
        except AttributeError:
            # pypy ctypes arrays do not feature a from_buffer() method.
            self._isshared = False
            # in case we requested a copy earlier, we do not need to recreate
            # the array, since we have it already. In any other case, create
            # a byte array.
            if not docopy:
                # Try to determine the itemsize again for array
                # instances, just in case the user assumed it to work.
                if isinstance(self._obj, _ARRAY.array):
                    itemsize = self._obj.itemsize
                    bsize = len(self._obj) * itemsize
                self._obj = create_array(self._obj, itemsize)
            self._view = (ctypes.c_ubyte * bsize)(*bytearray(self._obj))

    def __repr__(self):
        dtype = type(self._obj).__name__
        bsize = self.bytesize
        return "CTypesView(type=%s, bytesize=%d, shared=%s)" % (dtype, bsize,
                                                                self.is_shared)

    def __len__(self):
        """Returns the length of the underlying object in bytes."""
        return self.bytesize

    def to_bytes(self):
        """Casts the object to an array of bytes.
        
        If the view was created with ``docopy = False`` (the default), the
        returned object provides direct read-write access to the object data.
        Otherwise, the returned object will only modify a copy of the data.

        Returns:
            :obj:`ctypes._Pointer`: A pointer to a :class:`ctypes.c_uint8`
            array.

        """
        castval = ctypes.POINTER(ctypes.c_uint8 * self.bytesize)
        return ctypes.cast(self.view, castval).contents

    def to_uint16(self):
        """Casts the object to an array of 16-bit unsigned ints.
        
        If the view was created with ``docopy = False`` (the default), the
        returned object provides direct read-write access to the object data.
        Otherwise, the returned object will only modify a copy of the data.

        Returns:
            :obj:`ctypes._Pointer`: A pointer to a :class:`ctypes.c_uint16`
            array.

        """
        castval = ctypes.POINTER(ctypes.c_uint16 * (self.bytesize // 2))
        return ctypes.cast(self.view, castval).contents

    def to_uint32(self):
        """Casts the object to an array of 32-bit unsigned ints.
        
        If the view was created with ``docopy = False`` (the default), the
        returned object provides direct read-write access to the object data.
        Otherwise, the returned object will only modify a copy of the data.

        Returns:
            :obj:`ctypes._Pointer`: A pointer to a :class:`ctypes.c_uint32`
            array.

        """
        castval = ctypes.POINTER(ctypes.c_uint32 * (self.bytesize // 4))
        return ctypes.cast(self.view, castval).contents

    def to_uint64(self):
        """Casts the object to an array of 64-bit unsigned ints.
        
        If the view was created with ``docopy = False`` (the default), the
        returned object provides direct read-write access to the object data.
        Otherwise, the returned object will only modify a copy of the data.

        Returns:
            :obj:`ctypes._Pointer`: A pointer to a :class:`ctypes.c_uint64`
            array.

        """
        castval = ctypes.POINTER(ctypes.c_uint64 * (self.bytesize // 8))
        return ctypes.cast(self.view, castval).contents

    @property
    def bytesize(self):
        """int: The size (in bytes) of the underlying object."""
        return ctypes.sizeof(self.view)

    @property
    def view(self):
        """Provides a read/write-aware ``ctypes`` view of the encapsuled
        object.
        
        """
        return self._view

    @property
    def is_shared(self):
        """bool: Whether any modifications to the view will also modify the
        underlying object.

        """
        return self._isshared

    @property
    def object(self):
        """The underlying object."""
        return self._obj


class MemoryView(object):
    """A class that provides read-write access to indexable ``ctypes`` objects.

    .. note::
       ``MemoryView`` makes heavy use of recursion for multi-dimensional
       access, making it slow for many use-cases. For better performance, the
       ``numpy`` library can be used for fast access to many array-like data
       types, but it may support fewer types of arbitrary objects than the
       ``MemoryView`` class.

    Args:
        source: An arbitrary indexable ``ctypes`` object to access.
        itemsize (int): The size (in bytes) of each item of the indexable
            object.
        strides (tuple): The length of each dimension in the object (e.g.
            ``(4, 3)`` for a 4 x 3 2D array).
        getfunc (function, optional): Deprecated, do not use.
        setfunc (function, optional): Deprecated, do not use.
        srcsize (int, optional): The size (in bytes) of the input object.
            If ``len(source)`` returns the size of the object in bytes this will
            be inferred automatically, otherwise it must be manually specified.

    """
    def __init__(self, source, itemsize, strides, getfunc=None, setfunc=None,
                 srcsize=None):
        self._source = source
        self._itemsize = itemsize
        self._strides = strides
        self._srcsize = srcsize or len(source)
        self._offset = 0

        self._getfunc = getfunc or self._getbytes
        self._setfunc = setfunc or self._setbytes

        tsum = 1
        for v in strides:
            tsum *= v
        if tsum > self._srcsize:
            raise ValueError("strides exceed the accesible source size")
        #if itemsize > strides[-1]:
        #    raise ValueError("itemsize exceeds the accessible stride length")

    def _getbytes(self, start, end):
        """Gets the bytes within the range of start:end."""
        return self._source[start:end]

    def _setbytes(self, start, end, value):
        """Gets the bytes within the range of start:end to the passed
        value.
        """
        self._source[start:end] = value

    def _getindex(self, index):
        # Perform typechecking and preprocessing of view indices
        if type(index) is slice:
            raise IndexError("MemoryView slicing is not currently supported.")
        elif type(index) is not int:
            e = "Array indices must be integers (got '{0}')."
            raise TypeError(e.format(str(index)))
        else:
            final_idx = index
            if index < 0:
                final_idx = len(self) + index  # Handle negative indexing
            if index >= len(self):
                e = "Index {0} is out of bounds for a view of length {1}."
                raise IndexError(e.format(index, len(self)))
        return final_idx

    def __len__(self):
        """The length of the MemoryView over the current dimension
        (amount of items for the current dimension).
        """
        return self.strides[0]

    def __repr__(self):
        retval = "["
        slen = self.strides[0]
        for dim in range(slen - 1):
            retval += "%s, " % self[dim]
        retval += str(self[slen - 1])
        retval += "]"
        return retval

    def __getitem__(self, index):
        """Returns the item at the specified index."""
        index = self._getindex(index)
        if self.ndim == 1:
            offset = self._offset + index * self.itemsize
            return self._getfunc(offset, offset + self.itemsize)
        else:
            advance = self.itemsize
            for b in self.strides[1:]:
                advance *= b
            offset = self._offset + advance * index
            view = MemoryView(
                self._source, self.itemsize, self.strides[1:],
                self._getfunc, self._setfunc, self._srcsize
            )
            view._offset = offset
            return view

    def __setitem__(self, index, value):
        """Sets the item at index to the specified value."""
        index = self._getindex(index)
        offset = self._offset + index * self.itemsize
        if self.ndim == 1:
            self._setfunc(offset, offset + self.itemsize, value)
        else:
            advance = self.itemsize
            for b in self.strides[1:]:
                advance *= b
            offset = self._offset + advance * index
            view = MemoryView(
                self._source, self.itemsize, self.strides[1:],
                self._getfunc, self._setfunc, self._srcsize
            )
            view._offset = offset
            if len(value) != len(view):
                raise ValueError("value does not match the view strides")
            for x in range(len(view)):
                view[x] = value[x]

    @property
    def size(self):
        """int: The size (in bytes) of the underlying source object."""
        return self._srcsize

    @property
    def strides(self):
        """tuple: The length of each dimension the MemoryView."""
        return self._strides

    @property
    def itemsize(self):
        """int: The size (in bytes) of a single item in the object."""
        return self._itemsize

    @property
    def ndim(self):
        """int: The number of dimensions of the MemoryView."""
        return len(self.strides)

    @property
    def source(self):
        """The underlying data source."""
        return self._source



import ctypes
from .compat import isiterable, UnsupportedError
from .common import raise_sdl_err
from .array import to_ctypes
from .color import convert_to_color
from .. import surface, pixels, rect
from .algorithms import clipline
from .surface import _get_target_surface

__all__ = ["prepare_color", "fill", "line"]


def prepare_color(color, target):
    """Prepares a given color for a specific target.

    Targets can be :obj:`~sdl2.SDL_PixelFormat`, :obj:`~sdl2.SDL_Surface`,
    or :obj:`~sdl2.ext.SoftwareSprite` objects. 
    
    Colors can be provided in any form supported by
    :func:`sdl2.ext.convert_to_color`.
    
    Args:
        color (:obj:`sdl2.ext.Color`): The color to prepare for the pixel format
            of the given target.
        target (:obj:`SDL_PixelFormat`, :obj:`SDL_Surface`, :obj:`SoftwareSprite`): The
            target pixel format, surface, or sprite for which the color should be
            prepared.
    
    Returns:
        int: An integer approximating the given color in the target's pixel
        format.

    """
    color = convert_to_color(color)
    pformat = None
    # Software surfaces
    if isinstance(target, pixels.SDL_PixelFormat):
        pformat = target
    else:
        surf = _get_target_surface(target)
        pformat = surf.format.contents
    if pformat.Amask != 0:
        # Target has an alpha mask
        return pixels.SDL_MapRGBA(pformat, color.r, color.g, color.b, color.a)
    return pixels.SDL_MapRGB(pformat, color.r, color.g, color.b)


def fill(target, color, area=None):
    """Fills one or more rectangular areas on a surface with a given color.

    Fill areas can be specified as 4-item ``(x, y, w, h)`` tuples,
    :obj:`~sdl2.rect.SDL_Rect` objects, or a list containing multiple areas to
    fill in either format. If no area is provided, the entire target will be
    filled with the provided color.

    The fill color can be provided in any format supported by
    :func:`~sdl2.ext.color.convert_to_color`.

    Args:
        target (:obj:`~sdl2.SDL_Surface`, :obj:`~sdl2.ext.SoftwareSprite`): The
            target surface or sprite to modify.
        color (:obj:`sdl2.ext.Color`): The color with which to fill the
            specified region(s) of the target.
        area (tuple, :obj:`~sdl2.SDL_Rect`, list, optional): The rectangular
            region (or regions) of the target surface to fill with the given
            colour. If no regions are specified (the default), the whole surface
            of the target will be filled.

    """
    color = prepare_color(color, target)
    rtarget = _get_target_surface(target)

    err_msg = (
        "Fill areas must be specified as either (x, y, w, h) tuples or "
        "SDL_Rect objects (Got unsupported format '{0}')"
    )

    rects = []
    if area:
        if not isiterable(area) or not isiterable(area[0]):
            area = [area]
        for r in area:
            if isinstance(r, rect.SDL_Rect):
                rects.append(r)
            else:
                try:
                    new_rect = rect.SDL_Rect(
                        int(r[0]), int(r[1]), int(r[2]), int(r[3])
                    )
                    rects.append(new_rect)
                except (TypeError, ValueError, IndexError):
                    raise ValueError(err_msg.format(str(r)))

    if len(rects) > 2:
        rects, count = to_ctypes(rects, rect.SDL_Rect)
        rects = ctypes.cast(rects, ctypes.POINTER(rect.SDL_Rect))
        ret = surface.SDL_FillRects(rtarget, rects, count, color)
    elif len(rects) == 1:
        ret = surface.SDL_FillRect(rtarget, rects[0], color)
    else:
        ret = surface.SDL_FillRect(rtarget, None, color)
    if ret < 0:
        raise_sdl_err("filling the surface")


def line(target, color, dline, width=1):
    """Draws one or lines on a surface in a given color.

    The fill color can be provided in any format supported by
    :func:`~sdl2.ext.color.convert_to_color`.

    Args:
        target (:obj:`~sdl2.SDL_Surface`, :obj:`~sdl2.ext.SoftwareSprite`): The
            target surface or sprite to modify.
        color (:obj:`sdl2.ext.Color`): The color with which to draw lines.
        dline (tuple, list): The ``(x1, y1, x2, y2)`` integer coordinates of a
            line to draw, or a list of multiple sets of ``(x1, y1, x2, y2)``
            coordinates for multiple lines.
        width (int, optional): The width of the line(s) in pixels. Defaults to
            1 if not specified.

    """
    if width < 1:
        raise ValueError("width must be greater than 0")
    color = prepare_color(color, target)
    rtarget = _get_target_surface(target)

    # If first item is iterable, assume multiple lines in (x1, y1, x2, y2) form
    if isiterable(dline[0]):
        flattened = []
        for line in dline:
            flattened += list(line)
        dline = flattened

    # line: (x1, y1, x2, y2) OR (x1, y1, x2, y2, ...)
    if (len(dline) % 4) != 0:
        raise ValueError("line does not contain a valid set of points")
    pcount = len(dline)
    SDLRect = rect.SDL_Rect
    fillrect = surface.SDL_FillRect

    pitch = rtarget.pitch
    bpp = rtarget.format.contents.BytesPerPixel
    frac = pitch / bpp
    clip_rect = rtarget.clip_rect
    left, right = clip_rect.x, clip_rect.x + clip_rect.w - 1
    top, bottom = clip_rect.y, clip_rect.y + clip_rect.h - 1

    if bpp == 3:
        raise UnsupportedError("24bpp surfaces are not currently supported.")
    if bpp == 2:
        pxbuf = ctypes.cast(rtarget.pixels, ctypes.POINTER(ctypes.c_uint16))
    elif bpp == 4:
        pxbuf = ctypes.cast(rtarget.pixels, ctypes.POINTER(ctypes.c_uint32))
    else:
        pxbuf = rtarget.pixels  # byte-wise access.

    for idx in range(0, pcount, 4):
        x1, y1, x2, y2 = dline[idx:idx + 4]
        if x1 == x2:
            # Vertical line
            if y1 < y2:
                varea = SDLRect(x1 - width // 2, y1, width, y2 - y1)
            else:
                varea = SDLRect(x1 - width // 2, y2, width, y1 - y2)
            fillrect(rtarget, varea, color)
            continue
        if y1 == y2:
            # Horizontal line
            if x1 < x2:
                varea = SDLRect(x1, y1 - width // 2, x2 - x1, width)
            else:
                varea = SDLRect(x2, y1 - width // 2, x1 - x2, width)
            fillrect(rtarget, varea, color)
            continue
        if width != 1:
            raise ValueError("Diagonal lines must have a width of 1.")
        if width == 1:
            # Bresenham
            x1, y1, x2, y2 = clipline(left, top, right, bottom, x1, y1, x2, y2)
            x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
            if x1 is None:
                # not to be drawn
                continue
            dx = abs(x2 - x1)
            dy = -abs(y2 - y1)
            err = dx + dy
            sx, sy = 1, 1
            if x1 > x2:
                sx = -sx
            if y1 > y2:
                sy = -sy
            while True:
                pxbuf[int(y1 * frac + x1)] = color
                if x1 == x2 and y1 == y2:
                    break
                e2 = err * 2
                if e2 > dy:
                    err += dy
                    x1 += sx
                if e2 < dx:
                    err += dx
                    y1 += sy



import pytest
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_QuitSubSystem
from sdl2.error import SDL_GetError, SDL_ClearError
from sdl2 import touch


# Check if we have any touch devices before running tests
devices = 0
sdl2.SDL_SetHint(sdl2.SDL_HINT_MOUSE_TOUCH_EVENTS, b"1")
ret = sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO)
if ret == 0:
    devices = touch.SDL_GetNumTouchDevices()
SDL_Quit()

# Override the global fixture to enable trackpads as touch devices
@pytest.fixture(scope="module", autouse=True)
def with_sdl():
    sdl2.SDL_SetHint(sdl2.SDL_HINT_MOUSE_TOUCH_EVENTS, b"1")
    sdl2.SDL_ClearError()
    ret = sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    yield
    sdl2.SDL_Quit()


def test_SDL_GetNumTouchDevices():
    assert touch.SDL_GetNumTouchDevices() >= 0

@pytest.mark.skipif(devices == 0, reason="No available touch devices")
def test_SDL_GetTouchDevice():
    count = touch.SDL_GetNumTouchDevices()
    for i in range(count):
        dev_id = touch.SDL_GetTouchDevice(i)
        assert SDL_GetError() == b""
        assert dev_id != 0

@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
@pytest.mark.skipif(devices == 0, reason="No available touch devices")
def test_SDL_GetTouchDeviceType():
    types = [
        touch.SDL_TOUCH_DEVICE_INVALID, touch.SDL_TOUCH_DEVICE_DIRECT,
        touch.SDL_TOUCH_DEVICE_INDIRECT_ABSOLUTE,
        touch.SDL_TOUCH_DEVICE_INDIRECT_RELATIVE
    ]
    count = touch.SDL_GetNumTouchDevices()
    for i in range(count):
        dev_id = touch.SDL_GetTouchDevice(i)
        assert SDL_GetError() == b""
        assert dev_id != 0
        dev_type = touch.SDL_GetTouchDeviceType(dev_id)
        assert dev_type in types

@pytest.mark.skipif(devices == 0, reason="No available touch devices")
def test_SDL_GetNumTouchFingers():
    count = touch.SDL_GetNumTouchDevices()
    for i in range(count):
        dev_id = touch.SDL_GetTouchDevice(i)
        assert SDL_GetError() == b""
        assert dev_id != 0
        fingers = touch.SDL_GetNumTouchFingers(dev_id)
        assert SDL_GetError() == b""
        assert fingers >= 0

@pytest.mark.skipif(devices == 0, reason="No available touch devices")
def test_SDL_GetTouchFinger():
    count = touch.SDL_GetNumTouchDevices()
    for i in range(count):
        dev_id = touch.SDL_GetTouchDevice(i)
        assert SDL_GetError() == b""
        assert dev_id != 0
        fingers = touch.SDL_GetNumTouchFingers(dev_id)
        assert fingers >= 0
        for f in range(0, fingers):
            finger = touch.SDL_GetTouchFinger(dev_id, f)
            assert isinstance(finger.contents, touch.SDL_Finger)
            assert finger.contents.id >= 0
            assert 0 <= finger.contents.x <= 1
            assert 0 <= finger.contents.y <= 1
                


import sys
import pytest
import sdl2
from sdl2 import SDL_GetError
from sdl2 import clipboard
from sdl2.stdinc import SDL_TRUE, SDL_FALSE
from .conftest import SKIP_ANNOYING

@pytest.fixture
def window(with_sdl):
    flag = sdl2.SDL_WINDOW_BORDERLESS
    w = sdl2.SDL_CreateWindow(b"Test", 10, 40, 12, 13, flag)
    assert sdl2.SDL_GetError() == b""
    assert isinstance(w.contents, sdl2.SDL_Window)
    yield w
    sdl2.SDL_DestroyWindow(w)


@pytest.mark.skipif(SKIP_ANNOYING, reason="Skip unless requested")
def test_SDL_ClipboardText(window):
    # Test retrieving text from the clipboard
    ret = clipboard.SDL_GetClipboardText()
    original_contents = ret
    assert type(ret) in (str, bytes)
    # Test whether HasClipboardText is accurate
    expected = SDL_FALSE if len(ret) == 0 else SDL_TRUE
    assert clipboard.SDL_HasClipboardText() == expected
    # Set some new clipboard test and test for it
    sdl2.SDL_ClearError()
    ret = clipboard.SDL_SetClipboardText(b"test")
    assert SDL_GetError() == b""
    assert ret == 0
    assert clipboard.SDL_HasClipboardText() == SDL_TRUE
    assert clipboard.SDL_GetClipboardText() == b"test"
    # Reset original contents
    clipboard.SDL_SetClipboardText(original_contents)



import sys
import pytest
from ctypes import cast, c_char_p
import sdl2
from sdl2 import hints
from sdl2.stdinc import SDL_TRUE, SDL_FALSE

# Need to override global fixture to init/quit on every test
@pytest.fixture
def with_sdl():
    sdl2.SDL_ClearError()
    ret = sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    yield
    sdl2.SDL_Quit()


def test_SDL_ClearHints(with_sdl):
    assert hints.SDL_SetHint(b"TEST", b"32") == 1
    assert hints.SDL_GetHint(b"TEST") == b"32"
    hints.SDL_ClearHints()
    assert hints.SDL_GetHint(b"TEST") == None

def test_SDL_GetHint(with_sdl):
    assert hints.SDL_SetHint(b"TEST", b"32") == 1
    assert hints.SDL_GetHint(b"TEST") == b"32"
    assert hints.SDL_SetHint(hints.SDL_HINT_ALLOW_TOPMOST, b"true") == 1
    assert hints.SDL_GetHint(hints.SDL_HINT_ALLOW_TOPMOST) == b"true"

def test_SDL_SetHint(with_sdl):
    assert hints.SDL_SetHint(b"TEST", b"32") == 1
    assert hints.SDL_GetHint(b"TEST") == b"32"
    assert hints.SDL_SetHint(b"TEST", b"abcdef") == 1
    assert hints.SDL_GetHint(b"TEST") == b"abcdef"
    assert hints.SDL_SetHint(b"", b"hi") == 1
    assert hints.SDL_GetHint(b"") == b"hi"

def test_SDL_SetHintWithPriority(with_sdl):
    tst_hints = [
        [hints.SDL_HINT_DEFAULT, b"abc", b"abc"],
        [hints.SDL_HINT_NORMAL, b"ABC", b"ABC"],
        [hints.SDL_HINT_OVERRIDE, b"123", b"123"],
        [hints.SDL_HINT_NORMAL, b"xyz", b"123"],
    ]
    for priority, value, expected in tst_hints:
        ret = hints.SDL_SetHintWithPriority(b"TEST", value, priority)
        assert ret == (1 if value == expected else 0)
        assert hints.SDL_GetHint(b"TEST") == expected

def test_SDL_GetHintBoolean(with_sdl):
    assert hints.SDL_SetHint(b"TEST", b"32") == 1
    assert hints.SDL_GetHintBoolean(b"TEST", SDL_TRUE) == SDL_TRUE
    assert hints.SDL_GetHintBoolean(b"TEST", SDL_FALSE) == SDL_TRUE
    assert hints.SDL_GetHintBoolean(b"TEST2", SDL_FALSE) != SDL_TRUE

def test_SDL_AddDelHintCallback(with_sdl):
    topmost_hint = hints.SDL_HINT_ALLOW_TOPMOST
    calls = []
    def callback(userdata, name, oldval, newval):
        data = cast(userdata, c_char_p)
        calls.append((data.value, name, oldval, newval))
    hintcb = hints.SDL_HintCallback(callback)
    udata = c_char_p(b"banana")
    hints.SDL_AddHintCallback(topmost_hint, hintcb, udata)
    # SDL_AddHintCallback invokes the callback once.
    assert len(calls) == 1
    assert calls[0] == (b"banana", topmost_hint, None, None)
    hints.SDL_SetHint(topmost_hint, b"true")
    assert len(calls) == 2
    assert calls[1] == (b"banana", topmost_hint, None, b"true")
    hints.SDL_DelHintCallback(topmost_hint, hintcb, udata)
    hints.SDL_SetHint(topmost_hint, b"false")
    assert len(calls) == 2



import os
import sys
import ctypes
from ctypes.util import find_library
from ctypes import c_int, c_ubyte, c_float, byref, cast, POINTER, py_object
import pytest
import sdl2
from sdl2.stdinc import SDL_FALSE, SDL_TRUE, Uint16
from sdl2 import video, rect, pixels, surface, SDL_GetError
from .conftest import SKIP_ANNOYING

# Some tests don't work properly with some video drivers, so check the name
DRIVER_DUMMY = False
DRIVER_X11 = False
try:
    sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO)
    driver_name = video.SDL_GetCurrentVideoDriver()
    sdl2.SDL_Quit()
    DRIVER_DUMMY = driver_name == b"dummy"
    DRIVER_X11 = driver_name == b"x11"
except:
    pass

# Some tests don't work right on PyPy
is_pypy = hasattr(sys, "pypy_version_info")

if sys.version_info[0] >= 3:
    long = int

to_ctypes = lambda seq, dtype: (dtype * len(seq))(*seq)

def has_opengl_lib():
    for libname in("gl", "opengl", "opengl32"):
        path = find_library(libname)
        if path is not None:
            return True

def get_opengl_path():
    for libname in("gl", "opengl", "opengl32"):
        path = find_library(libname)
        if path is not None:
            return path

@pytest.fixture
def with_sdl_gl(with_sdl):
    ret = video.SDL_GL_LoadLibrary(None)
    assert SDL_GetError() == b""
    assert ret == 0
    yield
    video.SDL_GL_UnloadLibrary()

@pytest.fixture
def window(with_sdl):
    flag = video.SDL_WINDOW_BORDERLESS
    w = video.SDL_CreateWindow(b"Test", 10, 40, 12, 13, flag)
    assert SDL_GetError() == b""
    assert isinstance(w.contents, video.SDL_Window)
    yield w
    video.SDL_DestroyWindow(w)

@pytest.fixture
def decorated_window(with_sdl):
    flag = video.SDL_WINDOW_RESIZABLE
    w = video.SDL_CreateWindow(b"Test", 10, 40, 12, 13, flag)
    assert SDL_GetError() == b""
    assert isinstance(w.contents, video.SDL_Window)
    yield w
    video.SDL_DestroyWindow(w)

@pytest.fixture
def gl_window(with_sdl_gl):
    flag = video.SDL_WINDOW_OPENGL
    w = video.SDL_CreateWindow(b"OpenGL", 10, 40, 12, 13, flag)
    assert SDL_GetError() == b""
    ctx = video.SDL_GL_CreateContext(w)
    assert SDL_GetError() == b""
    yield (w, ctx)
    video.SDL_GL_DeleteContext(ctx)
    video.SDL_DestroyWindow(w)


# Test custom macros

def test_SDL_WINDOWPOS_UNDEFINED_DISPLAY():
    undef_mask = video.SDL_WINDOWPOS_UNDEFINED_MASK
    for x in range(0xFFFF):
        undef = video.SDL_WINDOWPOS_UNDEFINED_DISPLAY(x)
        assert undef_mask | x == undef
        assert (undef & undef_mask) == undef_mask
        assert undef != video.SDL_WINDOWPOS_CENTERED_DISPLAY(x)

def test_SDL_WINDOWPOS_ISUNDEFINED():
    assert video.SDL_WINDOWPOS_ISUNDEFINED(video.SDL_WINDOWPOS_UNDEFINED)
    assert not video.SDL_WINDOWPOS_ISUNDEFINED(video.SDL_WINDOWPOS_CENTERED)
    for x in range(0xFFFF):
        undef = video.SDL_WINDOWPOS_UNDEFINED_DISPLAY(x)
        assert video.SDL_WINDOWPOS_ISUNDEFINED(undef)

def test_SDL_WINDOWPOS_CENTERED_DISPLAY():
    centered_mask = video.SDL_WINDOWPOS_CENTERED_MASK
    for x in range(0xFFFF):
        centered = video.SDL_WINDOWPOS_CENTERED_DISPLAY(x)
        assert centered_mask | x == centered
        assert (centered & centered_mask) == centered_mask
        assert centered != video.SDL_WINDOWPOS_UNDEFINED_DISPLAY(x)

def test_SDL_WINDOWPOS_ISCENTERED():
    assert video.SDL_WINDOWPOS_ISCENTERED(video.SDL_WINDOWPOS_CENTERED)
    assert not video.SDL_WINDOWPOS_ISCENTERED(video.SDL_WINDOWPOS_UNDEFINED)
    for x in range(0xFFFF):
        centered = video.SDL_WINDOWPOS_CENTERED_DISPLAY(x)
        assert video.SDL_WINDOWPOS_ISCENTERED(centered)


# Test structures and classes

def test_SDL_Window():
    window = video.SDL_Window()
    assert isinstance(window, video.SDL_Window)


class TestSDLDisplayMode(object):

    def test_init(self):
        mode = video.SDL_DisplayMode()
        assert isinstance(mode, video.SDL_DisplayMode)
        fmt = sdl2.SDL_PIXELFORMAT_ARGB8888
        mode = video.SDL_DisplayMode(fmt, 800, 600, 60)
        assert isinstance(mode, video.SDL_DisplayMode)
        assert mode.format == fmt
        assert mode.w == 800
        assert mode.h == 600
        assert mode.refresh_rate == 60
        # Test exceptions on bad input
        with pytest.raises(TypeError):
            video.SDL_DisplayMode("Test")
        with pytest.raises(TypeError):
            video.SDL_DisplayMode(10, 10.6, 10, 10)
        with pytest.raises(TypeError):
            video.SDL_DisplayMode(10, 10, 10, None)

    def test___eq__(self):
        DMode = video.SDL_DisplayMode
        assert DMode() == DMode()
        assert DMode(10, 0, 0, 0) == DMode(10, 0, 0, 0)
        assert DMode(10, 10, 0, 0) == DMode(10, 10, 0, 0)
        assert DMode(10, 10, 10, 0) == DMode(10, 10, 10, 0)
        assert DMode(10, 10, 10, 10) == DMode(10, 10, 10, 10)
        assert DMode(0, 10, 0, 0) == DMode(0, 10, 0, 0)
        assert DMode(0, 0, 10, 0) == DMode(0, 0, 10, 0)
        assert DMode(0, 0, 0, 10) == DMode(0, 0, 0, 10)

        assert not (DMode() == DMode(10, 0, 0, 0))
        assert not (DMode(10, 0, 0, 0) == DMode(0, 0, 0, 0))
        assert not (DMode(10, 0, 0, 0) == DMode(0, 10, 0, 0))
        assert not (DMode(10, 0, 0, 0) == DMode(0, 0, 10, 0))
        assert not (DMode(10, 0, 0, 0) == DMode(0, 0, 0, 10))

    def test___ne__(self):
        DMode = video.SDL_DisplayMode
        assert not (DMode() != DMode())
        assert not (DMode(10, 0, 0, 0) != DMode(10, 0, 0, 0))
        assert not (DMode(10, 10, 0, 0) != DMode(10, 10, 0, 0))
        assert not (DMode(10, 10, 10, 0) != DMode(10, 10, 10, 0))
        assert not (DMode(10, 10, 10, 10) != DMode(10, 10, 10, 10))
        assert not (DMode(0, 10, 0, 0) != DMode(0, 10, 0, 0))
        assert not (DMode(0, 0, 10, 0) != DMode(0, 0, 10, 0))
        assert not (DMode(0, 0, 0, 10) != DMode(0, 0, 0, 10))

        assert DMode() != DMode(10, 0, 0, 0)
        assert DMode(10, 0, 0, 0) != DMode(0, 0, 0, 0)
        assert DMode(10, 0, 0, 0) != DMode(0, 10, 0, 0)
        assert DMode(10, 0, 0, 0) != DMode(0, 0, 10, 0)
        assert DMode(10, 0, 0, 0) != DMode(0, 0, 0, 10)


# Test module SDL functions

def test_SDL_VideoInitQuit():
    # Test with default driver
    assert sdl2.SDL_WasInit(0) & sdl2.SDL_INIT_VIDEO != sdl2.SDL_INIT_VIDEO
    ret = video.SDL_VideoInit(None)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0
    assert video.SDL_GetCurrentVideoDriver() # If initialized, should be string
    video.SDL_VideoQuit()
    assert not video.SDL_GetCurrentVideoDriver()
    # TODO: Test with string input (fails with b"dummy" for some reason?)

def test_SDL_GetNumVideoDrivers(with_sdl):
    numdrivers = video.SDL_GetNumVideoDrivers()
    assert numdrivers >= 1

def test_SDL_GetVideoDriver(with_sdl):
    numdrivers = video.SDL_GetNumVideoDrivers()
    for i in range(numdrivers):
        name = video.SDL_GetVideoDriver(i)
        assert type(name) in (str, bytes)

def test_SDL_GetCurrentVideoDriver(with_sdl):
    curdriver = video.SDL_GetCurrentVideoDriver()
    numdrivers = video.SDL_GetNumVideoDrivers()
    drivers = []
    for i in range(numdrivers):
        drivers.append(video.SDL_GetVideoDriver(i))
    assert curdriver in drivers

def test_SDL_GetNumVideoDisplays(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    assert numdisplays >= 1

def test_SDL_GetNumDisplayModes(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        modes = video.SDL_GetNumDisplayModes(index)
        assert modes >= 1

def test_SDL_GetDisplayMode(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        modes = video.SDL_GetNumDisplayModes(index)
        for mode in range(modes):
            dmode = video.SDL_DisplayMode()
            ret = video.SDL_GetDisplayMode(index, mode, byref(dmode))
            assert sdl2.SDL_GetError() == b""
            assert ret == 0
            if not DRIVER_DUMMY:
                assert dmode.w > 0
                assert dmode.h > 0

def test_SDL_GetCurrentDisplayMode(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        dmode = video.SDL_DisplayMode()
        ret = video.SDL_GetCurrentDisplayMode(index, byref(dmode))
        assert sdl2.SDL_GetError() == b""
        assert ret == 0
        assert dmode.w > 0
        assert dmode.h > 0

def test_SDL_GetDesktopDisplayMode(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        dmode = video.SDL_DisplayMode()
        ret = video.SDL_GetDesktopDisplayMode(index, byref(dmode))
        assert sdl2.SDL_GetError() == b""
        assert ret == 0
        assert dmode.w > 0
        assert dmode.h > 0

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GetClosestDisplayMode(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        dmode = video.SDL_DisplayMode()
        ret = video.SDL_GetCurrentDisplayMode(index, byref(dmode))
        assert sdl2.SDL_GetError() == b""
        assert ret == 0
        cmode = video.SDL_DisplayMode(
            dmode.format, dmode.w - 1, dmode.h - 1, dmode.refresh_rate
        )
        closest = video.SDL_DisplayMode()
        video.SDL_GetClosestDisplayMode(index, cmode, byref(closest))
        assert closest == dmode

def test_SDL_GetDisplayName(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        name = video.SDL_GetDisplayName(index)
        assert type(name) in (str, bytes)

def test_SDL_GetDisplayBounds(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        bounds = rect.SDL_Rect()
        ret = video.SDL_GetDisplayBounds(index, byref(bounds))
        assert sdl2.SDL_GetError() == b""
        assert ret == 0
        assert bounds.w > 0
        assert bounds.h > 0
        assert not rect.SDL_RectEmpty(bounds)

@pytest.mark.skipif(sdl2.dll.version < 2005, reason="not available")
def test_SDL_GetDisplayUsableBounds(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        bounds = rect.SDL_Rect()
        ret = video.SDL_GetDisplayUsableBounds(index, byref(bounds))
        assert ret == 0
        assert not rect.SDL_RectEmpty(bounds)

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GetDisplayDPI(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        ddpi, hdpi, vdpi = c_float(0), c_float(0), c_float(0)
        ret = video.SDL_GetDisplayDPI(
            index, byref(ddpi), byref(hdpi), byref(vdpi)
        )
        assert SDL_GetError() == b""
        assert ret == 0
        assert ddpi.value >= 96.0
        assert hdpi.value >= 96.0
        assert vdpi.value >= 96.0

@pytest.mark.skipif(sdl2.dll.version < 2009, reason="not available")
def test_SDL_GetDisplayOrientation(with_sdl):
    numdisplays = video.SDL_GetNumVideoDisplays()
    for index in range(numdisplays):
        orientation = video.SDL_GetDisplayOrientation(index)
        assert isinstance(orientation, int)
        assert orientation >= 0

def test_GetDisplayInfo(with_sdl):
    current = video.SDL_GetCurrentVideoDriver().decode('utf-8')
    print("Available Video Drivers:")
    for i in range(video.SDL_GetNumVideoDrivers()):
        name = video.SDL_GetVideoDriver(i).decode('utf-8')
        if name == current:
            name += " (*)"
        print(" - " + name)
    print("")
    print("Detected Displays:")
    for i in range(video.SDL_GetNumVideoDisplays()):
        name = video.SDL_GetDisplayName(i).decode('utf-8')
        info = " - " + name
        dm = video.SDL_DisplayMode()
        ret = video.SDL_GetDesktopDisplayMode(i, byref(dm))
        if ret == 0:
            res = " ({0}x{1} @ {2}Hz)".format(dm.w, dm.h, dm.refresh_rate)
            info += res
        print(info)

def test_SDL_CreateDestroyWindow(with_sdl):
    flag = video.SDL_WINDOW_BORDERLESS
    window = video.SDL_CreateWindow(b"Test", 10, 40, 12, 13, flag)
    assert SDL_GetError() == b""
    assert isinstance(window.contents, video.SDL_Window)
    video.SDL_DestroyWindow(window)

@pytest.mark.skip("not implemented")
def test_SDL_CreateWindowFrom(with_sdl):
    # No obvious cross-platform way to test this
    pass

def test_SDL_GetWindowDisplayIndex(window):
    numdisplays = video.SDL_GetNumVideoDisplays()
    dindex = video.SDL_GetWindowDisplayIndex(window)
    # Make sure display index is valid
    assert 0 <= dindex <= numdisplays

def test_SDL_GetWindowDisplayMode(window):
    # NOTE: Gets fullscreen mode of parent display, not size of window
    dmode = video.SDL_DisplayMode()
    ret = video.SDL_GetWindowDisplayMode(window, byref(dmode))
    assert SDL_GetError() == b""
    assert ret == 0
    assert dmode.w > 0
    assert dmode.h > 0

def test_SDL_SetWindowDisplayMode(window):
    # NOTE: Sets the fullscreen mode of the window, so can't easily test
    # NOTE: If set mode not supported, will change to closest supported res
    dindex = video.SDL_GetWindowDisplayIndex(window)
    dmode = video.SDL_DisplayMode()
    ret = video.SDL_GetCurrentDisplayMode(dindex, byref(dmode))
    assert ret == 0
    video.SDL_SetWindowDisplayMode(window, dmode)
    wmode = video.SDL_DisplayMode()
    ret = video.SDL_GetWindowDisplayMode(window, byref(wmode))
    assert SDL_GetError() == b""
    assert ret == 0
    assert dmode == wmode

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_GetWindowICCProfile(window):
    prof_size = ctypes.c_size_t(0)
    prof_ptr = video.SDL_GetWindowICCProfile(window, byref(prof_size))
    # This function returns a void pointer to the loaded ICC profile, which
    # needs to be cast to bytes to be read. As per the ICC spec, bytes
    # 36 to 39 of the header should always be 'acsp' in ASCII.
    if prof_size.value > 0:
        prof = ctypes.cast(prof_ptr, ctypes.POINTER(c_ubyte))
        assert bytes(prof[36:40]) == b"acsp"

def test_SDL_GetWindowPixelFormat(window):
    fmt = video.SDL_GetWindowPixelFormat(window)
    assert fmt in sdl2.ALL_PIXELFORMATS

def test_SDL_GetWindowID(window):
    assert video.SDL_GetWindowID(window) >= 0

def test_SDL_GetWindowFromID(window):
    window2 = video.SDL_GetWindowFromID(video.SDL_GetWindowID(window))
    assert video.SDL_GetWindowID(window) == video.SDL_GetWindowID(window2)
    assert video.SDL_GetWindowTitle(window) == video.SDL_GetWindowTitle(window2)
    # Make sure sizes/positions are the same
    px1, py1, px2, py2 = c_int(0), c_int(0), c_int(0), c_int(0)
    video.SDL_GetWindowPosition(window, byref(px1), byref(py1))
    video.SDL_GetWindowPosition(window2, byref(px2), byref(py2))
    assert (px1.value, py1.value) == (px2.value, py2.value)
    video.SDL_GetWindowSize(window, byref(px1), byref(py1))
    video.SDL_GetWindowSize(window2, byref(px2), byref(py2))
    assert (px1.value, py1.value) == (px2.value, py2.value)

def test_SDL_GetWindowFlags(with_sdl):
    flags = (
        video.SDL_WINDOW_BORDERLESS,
        video.SDL_WINDOW_BORDERLESS | video.SDL_WINDOW_HIDDEN,
        video.SDL_WINDOW_RESIZABLE
    )
    for flag in flags:
        win = video.SDL_CreateWindow(b"Test", 10, 10, 10, 10, flag)
        wflags = video.SDL_GetWindowFlags(win)
        assert (wflags & flag) == flag
        video.SDL_DestroyWindow(win)

def test_SDL_GetSetWindowTitle(window):
    assert video.SDL_GetWindowTitle(window) == b"Test"
    video.SDL_SetWindowTitle(window, b"Hello there")
    assert video.SDL_GetWindowTitle(window) == b"Hello there"

def test_SDL_SetWindowIcon(window):
    sf = surface.SDL_CreateRGBSurface(
        0, 16, 16, 16, 0xF000, 0x0F00, 0x00F0, 0x000F
    )
    assert isinstance(sf.contents, surface.SDL_Surface)
    video.SDL_SetWindowIcon(window, sf)
    assert SDL_GetError() == b""

@pytest.mark.xfail(is_pypy, reason="PyPy can't create proper py_object values")
def test_SDL_GetSetWindowData(window):
    values = {
        b"text": py_object("Teststring"),
        b"list": py_object([1, 2, "a", "b"]),
        b"tuple": py_object((1, 2, 3)),
    }
    for k, v in values.items():
        video.SDL_SetWindowData(window, k, v)
        retval = video.SDL_GetWindowData(window, k)
        assert retval.contents.value == v.value

@pytest.mark.xfail(DRIVER_X11, reason="Wonky with some window managers")
def test_SDL_GetSetWindowPosition(with_sdl):
    window = video.SDL_CreateWindow(b"Test", 10, 200, 10, 10, 0)
    px, py = c_int(0), c_int(0)
    video.SDL_GetWindowPosition(window, byref(px), byref(py))
    assert (px.value, py.value) == (10, 200)
    video.SDL_SetWindowPosition(window, 0, 150)
    video.SDL_GetWindowPosition(window, byref(px), byref(py))
    assert (px.value, py.value) == (0, 150)
    video.SDL_SetWindowPosition(window, 480, 320)
    video.SDL_GetWindowPosition(window, byref(px), byref(py))
    assert (px.value, py.value) == (480, 320)
    video.SDL_DestroyWindow(window)

def test_SDL_GetSetWindowSize(window):
    sx, sy = c_int(0), c_int(0)
    video.SDL_GetWindowSize(window, byref(sx), byref(sy))
    assert (sx.value, sy.value) == (12, 13)
    video.SDL_SetWindowSize(window, 1, 1)
    video.SDL_GetWindowSize(window, byref(sx), byref(sy))
    assert (sx.value, sy.value) == (1, 1)
    video.SDL_SetWindowSize(window, 480, 320)
    video.SDL_GetWindowSize(window, byref(sx), byref(sy))
    assert (sx.value, sy.value) == (480, 320)
    # Test that negative sizes are ignored
    video.SDL_SetWindowSize(window, -200, -10)
    video.SDL_GetWindowSize(window, byref(sx), byref(sy))
    assert (sx.value, sy.value) == (480, 320)

def test_SDL_GetWindowBordersSize(window, decorated_window):
    # Currently, only X11 and Windows video drivers support border size
    supports_borders = video.SDL_GetCurrentVideoDriver() in [b"x11", b"windows"]
    # For a decorated window, make sure all borders are >= 0
    video.SDL_ShowWindow(decorated_window)
    t, l, b, r = c_int(), c_int(), c_int(), c_int()
    ret = video.SDL_GetWindowBordersSize(
       decorated_window, byref(t), byref(l), byref(b), byref(r)
    )
    values = [x.value for x in (t, l, b, r)]
    assert all([v >= 0 for v in values])
    if supports_borders:
        assert ret == 0
    # Test again with a borderless window & make sure borders are all 0
    video.SDL_ShowWindow(window)
    ret = video.SDL_GetWindowBordersSize(
        window, byref(t), byref(l), byref(b), byref(r)
    )
    values = [x.value for x in (t, l, b, r)]
    assert all([v == 0 for v in values])
    if supports_borders:
        assert ret == 0

def test_SDL_GetSetWindowMinimumSize(window):
    sx, sy = c_int(0), c_int(0)
    minx, miny = c_int(0), c_int(0)
    video.SDL_GetWindowSize(window, byref(sx), byref(sy))
    assert (sx.value, sy.value) == (12, 13)
    # Set and verify the minimum window size
    video.SDL_SetWindowMinimumSize(window, 10, 10)
    assert SDL_GetError() == b""
    video.SDL_GetWindowMinimumSize(window, byref(minx), byref(miny))
    assert (minx.value, miny.value) == (10, 10)
    # Make sure window can't be set below its minimum size
    video.SDL_SetWindowSize(window, 1, 1)
    video.SDL_GetWindowSize(window, byref(sx), byref(sy))
    assert (sx.value, sy.value) == (10, 10)

def test_SDL_GetSetWindowMaximumSize(window):
    sx, sy = c_int(0), c_int(0)
    maxx, maxy = c_int(0), c_int(0)
    video.SDL_GetWindowSize(window, byref(sx), byref(sy))
    assert (sx.value, sy.value) == (12, 13)
    # Set and verify the maximum window size
    video.SDL_SetWindowMaximumSize(window, 32, 32)
    assert SDL_GetError() == b""
    video.SDL_GetWindowMaximumSize(window, byref(maxx), byref(maxy))
    assert (maxx.value, maxy.value) == (32, 32)
    # Make sure window can't be set above its maximum size
    video.SDL_SetWindowSize(window, 50, 50)
    video.SDL_GetWindowSize(window, byref(sx), byref(sy))
    assert (sx.value, sy.value) == (32, 32)

def test_SDL_SetWindowBordered(window):
    border_flag = video.SDL_WINDOW_BORDERLESS
    assert video.SDL_GetWindowFlags(window) & border_flag == border_flag
    video.SDL_SetWindowBordered(window, SDL_TRUE)
    if not DRIVER_DUMMY:
        assert not video.SDL_GetWindowFlags(window) & border_flag == border_flag
        video.SDL_SetWindowBordered(window, SDL_FALSE)
        assert video.SDL_GetWindowFlags(window) & border_flag == border_flag

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_SetWindowResizable(window):
    flags = video.SDL_GetWindowFlags(window)
    assert flags & video.SDL_WINDOW_RESIZABLE != video.SDL_WINDOW_RESIZABLE
    video.SDL_SetWindowResizable(window, SDL_TRUE)
    flags = video.SDL_GetWindowFlags(window)
    assert flags & video.SDL_WINDOW_RESIZABLE == video.SDL_WINDOW_RESIZABLE
    video.SDL_SetWindowResizable(window, SDL_FALSE)
    flags = video.SDL_GetWindowFlags(window)
    assert flags & video.SDL_WINDOW_RESIZABLE != video.SDL_WINDOW_RESIZABLE

@pytest.mark.skip("Test doesn't work, may need to be interactive")
@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_SetWindowAlwaysOnTop(with_sdl):
    ON_TOP_FLAG = video.SDL_WINDOW_ALWAYS_ON_TOP
    window = video.SDL_CreateWindow(
        b"Always On Top", 10, 100, 10, 10, ON_TOP_FLAG
    )
    video.SDL_ShowWindow(window)
    flags = video.SDL_GetWindowFlags(window)
    assert flags & ON_TOP_FLAG == ON_TOP_FLAG
    video.SDL_SetWindowAlwaysOnTop(window, SDL_FALSE)
    flags = video.SDL_GetWindowFlags(window)
    assert flags & ON_TOP_FLAG != ON_TOP_FLAG
    video.SDL_SetWindowAlwaysOnTop(window, SDL_TRUE)
    flags = video.SDL_GetWindowFlags(window)
    assert flags & ON_TOP_FLAG == ON_TOP_FLAG
    video.SDL_DestroyWindow(window)

def test_SDL_ShowHideWindow(window):
    shown_flag = video.SDL_WINDOW_SHOWN
    video.SDL_ShowWindow(window)
    assert video.SDL_GetWindowFlags(window) & shown_flag == shown_flag
    video.SDL_HideWindow(window)
    assert not video.SDL_GetWindowFlags(window) & shown_flag == shown_flag

def test_SDL_RaiseWindow(window):
    # NOTE: Doesn't set any flags, so can't test this super well
    video.SDL_RaiseWindow(window)

@pytest.mark.skipif(SKIP_ANNOYING, reason="Skip unless requested")
def test_SDL_MaximizeWindow(decorated_window):
    shown_flag = video.SDL_WINDOW_SHOWN
    max_flag = video.SDL_WINDOW_MAXIMIZED
    window = decorated_window
    video.SDL_ShowWindow(window)
    assert video.SDL_GetWindowFlags(window) & shown_flag == shown_flag
    assert not video.SDL_GetWindowFlags(window) & max_flag == max_flag
    video.SDL_MaximizeWindow(window)
    if not DRIVER_DUMMY:
        assert video.SDL_GetWindowFlags(window) & max_flag == max_flag

@pytest.mark.skipif(SKIP_ANNOYING, reason="Skip unless requested")
def test_SDL_MinimizeRestoreWindow(decorated_window):
    shown_flag = video.SDL_WINDOW_SHOWN
    min_flag = video.SDL_WINDOW_MINIMIZED
    window = decorated_window
    video.SDL_ShowWindow(window)
    assert video.SDL_GetWindowFlags(window) & shown_flag == shown_flag
    assert not video.SDL_GetWindowFlags(window) & min_flag == min_flag
    video.SDL_MinimizeWindow(window)
    if not (DRIVER_DUMMY or DRIVER_X11):
        assert video.SDL_GetWindowFlags(window) & min_flag == min_flag
    video.SDL_RestoreWindow(window)
    if not (DRIVER_DUMMY or DRIVER_X11):
        assert not video.SDL_GetWindowFlags(window) & min_flag == min_flag

def test_SDL_SetWindowFullscreen(with_sdl):
    # TODO: Add non-hidden test once annoying test toggle implemented
    flags = (
        video.SDL_WINDOW_BORDERLESS | video.SDL_WINDOW_HIDDEN,
        video.SDL_WINDOW_RESIZABLE | video.SDL_WINDOW_HIDDEN,
    )
    is_fullscreen = video.SDL_WINDOW_FULLSCREEN
    for flag in flags:
        window = video.SDL_CreateWindow(b"Test", 0, 0, 1024, 768, flag)
        video.SDL_SetWindowFullscreen(window, True)
        flags = video.SDL_GetWindowFlags(window)
        assert flags & is_fullscreen == is_fullscreen
        video.SDL_SetWindowFullscreen(window, False)
        flags = video.SDL_GetWindowFlags(window)
        assert flags & is_fullscreen != is_fullscreen
        video.SDL_DestroyWindow(window)

def test_SDL_GetWindowSurface(window):
    sf = video.SDL_GetWindowSurface(window)
    assert SDL_GetError() == b""
    assert isinstance(sf.contents, surface.SDL_Surface)

def test_SDL_UpdateWindowSurface(window):
    sf = video.SDL_GetWindowSurface(window)
    assert isinstance(sf.contents, surface.SDL_Surface)
    ret = video.SDL_UpdateWindowSurface(window)
    assert SDL_GetError() == b""
    assert ret == 0

def test_SDL_UpdateWindowSurfaceRects(window):
    sf = video.SDL_GetWindowSurface(window)
    assert isinstance(sf.contents, surface.SDL_Surface)
    rectlist = (rect.SDL_Rect * 4)(
        rect.SDL_Rect(0, 0, 0, 0),
        rect.SDL_Rect(10, 10, 10, 10),
        rect.SDL_Rect(0, 0, 5, 4),
        rect.SDL_Rect(-5, -5, 6, 2)
    )
    rect_ptr = cast(rectlist, POINTER(rect.SDL_Rect))
    ret = video.SDL_UpdateWindowSurfaceRects(window, rect_ptr, 4)
    assert SDL_GetError() == b""
    assert ret == 0

@pytest.mark.skip("Can't set window grab for some reason")
def test_SDL_GetSetWindowGrab(decorated_window):
    window = decorated_window
    video.SDL_ShowWindow(window)
    assert video.SDL_GetWindowGrab(window) == SDL_FALSE
    video.SDL_SetWindowGrab(window, SDL_TRUE)
    assert video.SDL_GetWindowGrab(window) == SDL_TRUE
    video.SDL_SetWindowGrab(window, SDL_FALSE)
    assert video.SDL_GetWindowGrab(window) == SDL_FALSE

@pytest.mark.skip("Can't set window grab for some reason")
@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_GetSetWindowKeyboardGrab(decorated_window):
    window = decorated_window
    video.SDL_ShowWindow(window)
    assert video.SDL_GetWindowKeyboardGrab(window) == SDL_FALSE
    video.SDL_SetWindowKeyboardGrab(window, SDL_TRUE)
    assert video.SDL_GetWindowKeyboardGrab(window) == SDL_TRUE
    video.SDL_SetWindowKeyboardGrab(window, SDL_FALSE)
    assert video.SDL_GetWindowKeyboardGrab(window) == SDL_FALSE

@pytest.mark.skip("Can't set window grab for some reason")
@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_GetSetWindowMouseGrab(decorated_window):
    window = decorated_window
    video.SDL_ShowWindow(window)
    assert video.SDL_GetWindowMouseGrab(window) == SDL_FALSE
    video.SDL_SetWindowMouseGrab(window, SDL_TRUE)
    assert video.SDL_GetWindowMouseGrab(window) == SDL_TRUE
    video.SDL_SetWindowMouseGrab(window, SDL_FALSE)
    assert video.SDL_GetWindowMouseGrab(window) == SDL_FALSE

@pytest.mark.skip("not implemented")
def test_SDL_GetGrabbedWindow(window):
    # NOTE: Should implement this once the above tests are fixed
    pass

@pytest.mark.skipif(sdl2.dll.version < 2018, reason="not available")
def test_SDL_GetSetWindowMouseRect(with_sdl):
    flags = video.SDL_WINDOW_BORDERLESS
    bounds_in = rect.SDL_Rect(0, 0, 100, 50)
    window = video.SDL_CreateWindow(b"Test", 200, 200, 200, 200, flags)
    # Try setting a mouse boundary
    ret = video.SDL_SetWindowMouseRect(window, byref(bounds_in))
    assert SDL_GetError() == b""
    assert ret == 0
    bounds_out = video.SDL_GetWindowMouseRect(window)
    assert bounds_out != None
    assert bounds_in == bounds_out.contents
    # Try removing the boundary
    ret = video.SDL_SetWindowMouseRect(window, None)
    assert SDL_GetError() == b""
    assert ret == 0
    bounds_out = video.SDL_GetWindowMouseRect(window)
    assert not bounds_out  # bounds_out should be null pointer
    video.SDL_DestroyWindow(window)

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GetSetWindowBrightness(window):
    orig = video.SDL_GetWindowBrightness(window)
    assert isinstance(orig, float)
    assert orig >= 0
    # Go from 0.0, 0.1 ... to 1.0
    gammas = (x * 0.1 for x in range(0, 10))
    count = 0
    for b in gammas:
        ret = video.SDL_SetWindowBrightness(window, b)
        if ret == 0:
            val = video.SDL_GetWindowBrightness(window)
            assert round(abs(val-b), 7) == 0
            count += 1
    assert count > 0

def test_SDL_GetSetWindowOpacity(window):
    opacity = c_float(0)
    ret = video.SDL_GetWindowOpacity(window, byref(opacity))
    assert ret == 0
    assert opacity.value == 1.0
    if not DRIVER_DUMMY:
        ret = video.SDL_SetWindowOpacity(window, 0.5)
        assert SDL_GetError() == b""
        assert ret == 0
        ret = video.SDL_GetWindowOpacity(window, byref(opacity))
        assert SDL_GetError() == b""
        assert ret == 0
        assert opacity.value == 0.5

@pytest.mark.skipif(sdl2.dll.version < 2005, reason="not available")
def test_SDL_SetWindowModalFor(window, decorated_window):
    # NOTE: Only supported on X11
    ret = video.SDL_SetWindowModalFor(window, decorated_window)
    if video.SDL_GetCurrentVideoDriver() == b"x11":
        assert SDL_GetError() == b""
        assert ret == 0

@pytest.mark.skipif(sdl2.dll.version < 2005, reason="not available")
def test_SDL_SetWindowInputFocus(window):
    # NOTE: Only supported on X11 and Wayland
    ret = video.SDL_SetWindowInputFocus(window)
    if video.SDL_GetCurrentVideoDriver() in [b"x11", b"wayland"]:
        assert SDL_GetError() == b""
        assert ret == 0

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GetSetWindowGammaRamp(window):
    vals = (Uint16 * 256)()
    pixels.SDL_CalculateGammaRamp(0.5, vals)
    ret = video.SDL_SetWindowGammaRamp(window, vals, vals, vals)
    assert SDL_GetError() == b""
    assert ret == 0
    r = (Uint16 * 256)()
    g = (Uint16 * 256)()
    b = (Uint16 * 256)()
    ret = video.SDL_GetWindowGammaRamp(window, r, g, b)
    assert SDL_GetError() == b""
    assert ret == 0
    for i in range(len(vals)):
        assert r[i] == vals[i]
        assert g[i] == vals[i]
        assert b[i] == vals[i]

@pytest.mark.skip("not implemented")
@pytest.mark.skipif(sdl2.dll.version < 2004, reason="not available")
def test_SDL_SetWindowHitTest(self):
    # NOTE: This sets a callback that's triggered when you mouse over certain
    # regions of a window. Pretty sure this can't be tested non-interactively.
    pass

@pytest.mark.skipif(sdl2.dll.version < 2016, reason="not available")
def test_SDL_FlashWindow(window):
    # NOTE: Not the most comprehensive test, but it does test the basic bindings
    ret = video.SDL_FlashWindow(window, video.SDL_FLASH_BRIEFLY)
    if not DRIVER_DUMMY:
        assert SDL_GetError() == b""
        assert ret == 0

def test_screensaver(with_sdl):
    video.SDL_EnableScreenSaver()
    assert video.SDL_IsScreenSaverEnabled() == SDL_TRUE
    video.SDL_EnableScreenSaver()
    assert video.SDL_IsScreenSaverEnabled() == SDL_TRUE
    video.SDL_DisableScreenSaver()
    assert video.SDL_IsScreenSaverEnabled() == SDL_FALSE
    video.SDL_DisableScreenSaver()
    assert video.SDL_IsScreenSaverEnabled() == SDL_FALSE
    video.SDL_EnableScreenSaver()
    assert video.SDL_IsScreenSaverEnabled() == SDL_TRUE
    video.SDL_DisableScreenSaver()
    assert video.SDL_IsScreenSaverEnabled() == SDL_FALSE


# Test SDL OpenGL functions

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GL_LoadUnloadLibrary(with_sdl):
    # TODO: Test whether other GL functions work after GL is unloaded
    # (unloading doesn't always work right on macOS for some reason)
    ret = video.SDL_GL_LoadLibrary(None)
    assert SDL_GetError() == b""
    assert ret == 0
    video.SDL_GL_UnloadLibrary()
    # Try loading a library from a path
    if has_opengl_lib():
        fpath = get_opengl_path().encode("utf-8")
        ret = video.SDL_GL_LoadLibrary(fpath)
        assert SDL_GetError() == b""
        assert ret == 0
        video.SDL_GL_UnloadLibrary()

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GL_CreateDeleteContext(with_sdl_gl):
    window = video.SDL_CreateWindow(
        b"OpenGL", 10, 40, 32, 24, video.SDL_WINDOW_OPENGL
    )
    ctx = video.SDL_GL_CreateContext(window)
    assert SDL_GetError() == b""
    video.SDL_GL_DeleteContext(ctx)
    ctx = video.SDL_GL_CreateContext(window)
    assert SDL_GetError() == b""
    video.SDL_GL_DeleteContext(ctx)
    video.SDL_DestroyWindow(window)

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GL_GetProcAddress(gl_window):
    procaddr = video.SDL_GL_GetProcAddress(b"glGetString")
    assert SDL_GetError() == b""
    assert procaddr is not None and int(procaddr) != 0

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GL_ExtensionSupported(gl_window):
    assert video.SDL_GL_ExtensionSupported(b"GL_EXT_bgra")
    assert SDL_GetError() == b""

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GL_GetSetResetAttribute(with_sdl_gl):
    # Create a context and get its bit depth
    window = video.SDL_CreateWindow(
        b"OpenGL", 10, 40, 12, 13, video.SDL_WINDOW_OPENGL
    )
    ctx = video.SDL_GL_CreateContext(window)
    bufstate = c_int(0)
    ret = video.SDL_GL_GetAttribute(video.SDL_GL_DOUBLEBUFFER, byref(bufstate))
    video.SDL_GL_DeleteContext(ctx)
    video.SDL_DestroyWindow(window)
    assert SDL_GetError() == b""
    assert ret == 0
    # Try setting a different GL bit depth
    new_bufstate = 0 if bufstate.value == 1 else 1
    video.SDL_GL_SetAttribute(video.SDL_GL_DOUBLEBUFFER, new_bufstate)
    assert SDL_GetError() == b""
    assert ret == 0
    # Create a new context to see if it's using the new bit depth 
    window = video.SDL_CreateWindow(
        b"OpenGL", 10, 40, 12, 13, video.SDL_WINDOW_OPENGL
    )
    ctx = video.SDL_GL_CreateContext(window)
    val = c_int(0)
    ret = video.SDL_GL_GetAttribute(video.SDL_GL_DOUBLEBUFFER, byref(val))
    video.SDL_GL_DeleteContext(ctx)
    video.SDL_DestroyWindow(window)
    assert SDL_GetError() == b""
    assert ret == 0
    assert bufstate.value != val.value
    assert val.value == new_bufstate
    # Try resetting the context and see if it goes back to the original depth
    video.SDL_GL_ResetAttributes()
    window = video.SDL_CreateWindow(
        b"OpenGL", 10, 40, 12, 13, video.SDL_WINDOW_OPENGL
    )
    ctx = video.SDL_GL_CreateContext(window)
    val = c_int(0)
    ret = video.SDL_GL_GetAttribute(video.SDL_GL_DOUBLEBUFFER, byref(val))
    video.SDL_GL_DeleteContext(ctx)
    video.SDL_DestroyWindow(window)
    assert bufstate.value == val.value

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GL_MakeCurrent(gl_window):
    window, ctx = gl_window
    ret = video.SDL_GL_MakeCurrent(window, ctx)
    assert SDL_GetError() == b""
    assert ret == 0

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GL_GetSetSwapInterval(gl_window):
    window, ctx = gl_window
    ret = video.SDL_GL_MakeCurrent(window, ctx)
    assert SDL_GetError() == b""
    assert ret == 0
    # Try enabling/disabling OpenGL vsync
    for value in [0, 1]:
        ret = video.SDL_GL_SetSwapInterval(value)
        if ret == 0:
            assert video.SDL_GL_GetSwapInterval() == value

@pytest.mark.skipif(DRIVER_DUMMY, reason="Doesn't work with dummy driver")
def test_SDL_GL_SwapWindow(gl_window):
    window, ctx = gl_window
    ret = video.SDL_GL_MakeCurrent(window, ctx)
    assert SDL_GetError() == b""
    assert ret == 0
    video.SDL_GL_SwapWindow(window)
    video.SDL_GL_SwapWindow(window)
    video.SDL_GL_SwapWindow(window)
    assert SDL_GetError() == b""



# Test font rendering using the FontTTF class

# This example allows you to try typing out some text in a given font, and
# supports the following keys:
#   - Left/Right Arrow Keys: Cycles text alignment (left/right/center)
#   - Up/Down Arrow Keys: Increases/decreases line height
#   - Tab: Cycles font rendering style

import os
import sys
import sdl2.ext
from sdl2.sdlttf import TTF_FontLineSkip
from sdl2.ext import FontTTF

filepath = os.path.abspath(os.path.dirname(__file__))
RESOURCES = sdl2.ext.Resources(filepath, "resources")
BLACK_RGBA = (0, 0, 0, 255)
WHITE_RGBA = (255, 255, 255, 255)


def update_text(renderer, surface):
    # Create a texture for the surface and render it to the screen
    tx = sdl2.ext.Texture(renderer, surface)
    renderer.clear(BLACK_RGBA)
    renderer.copy(tx, dstrect=(10, 10))
    renderer.present()


def run():
    # Initialize SDL2 and create a Window and Renderer
    sdl2.ext.init()
    window = sdl2.ext.Window("Font Rendering", size=(800, 500))
    renderflags = sdl2.SDL_RENDERER_SOFTWARE
    if "-hardware" in sys.argv:
        renderflags = (
            sdl2.SDL_RENDERER_ACCELERATED | sdl2.SDL_RENDERER_PRESENTVSYNC
        )
    renderer = sdl2.ext.Renderer(window, flags=renderflags)
    window.show()

    # Create and initialize a font to render text with
    fontpath = RESOURCES.get_path("tuffy.ttf")
    font = FontTTF(fontpath, "20px", WHITE_RGBA)

    # Add some additional font styles
    styles = ['default', 'small', 'red', 'large', 'bg_fill']
    font.add_style('small', '10px', WHITE_RGBA)
    font.add_style('red', '20px', (255, 0, 0))
    font.add_style('large', '35px', WHITE_RGBA)
    font.add_style('bg_fill', '20px', BLACK_RGBA, WHITE_RGBA)

    # Initialize font rendering options
    line_height = TTF_FontLineSkip(font.get_ttf_font())
    alignments = ["left", "center", "right"]
    align_idx = 0
    style_idx = 0

    # Set a default string with which to render text
    txt = u"Hi There!\nYou can edit this text using the keyboard and delete keys."

    # Render the text and present it on the screen
    txt_rendered = font.render_text(txt, width=780)
    update_text(renderer, txt_rendered)

    # Tell SDL2 to start reading Text Editing events. This allows for proper
    # handling of unicode characters and modifier keys.
    sdl2.SDL_StartTextInput()

    # Create a simple event loop and wait for keydown, text editing, and quit events.
    running = True
    while running:
        events = sdl2.ext.get_events()
        for event in events:
            update_txt = False
            if event.type == sdl2.SDL_QUIT:
                running = False
                break

            # Handle non-standard keyboard events
            elif event.type == sdl2.SDL_KEYDOWN:
                update_txt = True
                sdl_keysym = event.key.keysym.sym
                # If backspace pressed, remove last character (if any) from txt
                if sdl_keysym == sdl2.SDLK_BACKSPACE:
                    txt = txt[:-1]
                # If enter/return pressed, insert a newline
                elif sdl_keysym == sdl2.SDLK_RETURN:
                    txt = txt + u"\n"
                # If left or right arrow pressed, change text alignment mode
                elif sdl_keysym == sdl2.SDLK_LEFT:
                    align_idx = (align_idx - 1) % 3
                elif sdl_keysym == sdl2.SDLK_RIGHT:
                    align_idx = (align_idx + 1) % 3
                elif sdl_keysym == sdl2.SDLK_UP:
                    line_height += 1
                elif sdl_keysym == sdl2.SDLK_DOWN:
                    if line_height > 1:
                        line_height -= 1 
                # If tab pressed, cycle through the different font styles
                elif sdl_keysym == sdl2.SDLK_TAB:
                    style_idx = (style_idx + 1) % len(styles)

            # Handle text input events
            elif event.type == sdl2.SDL_TEXTINPUT:
                update_txt = True
                txt += event.text.text.decode("utf-8")

            # If txt has changed since the start of the loop, update the renderer
            if update_txt:
                align = alignments[align_idx]
                style = styles[style_idx]
                txt_rendered = font.render_text(
                    txt, style, width=780, line_h=line_height, align=align
                )
                update_text(renderer, txt_rendered)

    # Now that we're done, close everything down and quit SDL2
    font.close()
    renderer.destroy()
    window.close()
    sdl2.ext.quit()
    return 0

if __name__ == "__main__":
    sys.exit(run())



import abc
from ctypes import byref, c_int

from .. import surface, rect, render
from ..stdinc import Uint32

from .common import SDLError
from .surface import subsurface

__all__ = ["Sprite", "SoftwareSprite", "TextureSprite"]


class Sprite(object):
    """A simple 2D object."""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Creates a new Sprite."""
        super(Sprite, self).__init__()
        self.x = 0
        self.y = 0
        self.depth = 0

    @property
    def position(self):
        """The top-left position of the Sprite as tuple."""
        return self.x, self.y

    @position.setter
    def position(self, value):
        """The top-left position of the Sprite as tuple."""
        self.x = value[0]
        self.y = value[1]

    @property
    @abc.abstractmethod
    def size(self):
        """The size of the Sprite as tuple."""
        return

    @property
    def area(self):
        """The rectangular area occupied by the Sprite."""
        w, h = self.size
        return (self.x, self.y, self.x + w, self.y + h)


class SoftwareSprite(Sprite):
    """A simple, visible, pixel-based 2D object using software buffers."""
    def __init__(self, imgsurface, free):
        """Creates a new SoftwareSprite."""
        super(SoftwareSprite, self).__init__()
        self.free = free
        if isinstance(imgsurface, surface.SDL_Surface):
            self.surface = imgsurface
        elif "SDL_Surface" in str(type(imgsurface)):
            self.surface = imgsurface.contents
        else:
            raise TypeError("imgsurface must be an SDL_Surface")

    def __del__(self):
        """Releases the bound SDL_Surface, if it was created by the
        SoftwareSprite.
        """
        imgsurface = getattr(self, "surface", None)
        if self.free and imgsurface is not None:
            surface.SDL_FreeSurface(imgsurface)
        self.surface = None

    @property
    def size(self):
        """The size of the SoftwareSprite as tuple."""
        return self.surface.w, self.surface.h

    def subsprite(self, area):
        """Creates another SoftwareSprite from a part of the SoftwareSprite.

        The two sprites share pixel data, so if the parent sprite's surface is
        not managed by the sprite (free is False), you will need to keep it
        alive while the subsprite exists."""
        ssurface = subsurface(self.surface, area)
        ssprite = SoftwareSprite(ssurface, True)
        # Keeps the parent surface alive until subsprite is freed
        if self.free:
            ssprite._parent = self
        return ssprite

    def __repr__(self):
        return "SoftwareSprite(size=%s, bpp=%d)" % \
            (self.size, self.surface.format.contents.BitsPerPixel)


class TextureSprite(Sprite):
    """A simple, visible, texture-based 2D object, using a renderer."""
    def __init__(self, texture, free=True):
        """Creates a new TextureSprite."""
        super(TextureSprite, self).__init__()
        self.texture = texture
        flags = Uint32()
        access = c_int()
        w = c_int()
        h = c_int()
        ret = render.SDL_QueryTexture(texture, byref(flags), byref(access),
                                      byref(w), byref(h))
        if ret == -1:
            raise SDLError()
        self.free = free
        self.angle = 0.0
        self.flip = render.SDL_FLIP_NONE
        self._size = w.value, h.value
        self._center = None

    def __del__(self):
        """Releases the bound SDL_Texture."""
        if self.free:
            if self.texture != None:
                render.SDL_DestroyTexture(self.texture)
        self.texture = None

    @property
    def center(self):
        """The center of the TextureSprite as tuple."""
        return self._center

    @center.setter
    def center(self, value):
        """Sets the center of the TextureSprite."""
        if value != None:
            self._center = rect.SDL_Point(value[0], value[1])
        else:
            self._center = None

    @property
    def size(self):
        """The size of the TextureSprite as tuple."""
        return self._size

    def __repr__(self):
        flags = Uint32()
        access = c_int()
        w = c_int()
        h = c_int()
        ret = render.SDL_QueryTexture(self.texture, byref(flags),
                                      byref(access), byref(w), byref(h))
        if ret == -1:
            raise SDLError()
        if self.center:
            return "TextureSprite(format=%d, access=%d, size=%s, angle=%f, center=%s)" % \
                (flags.value, access.value, (w.value, h.value), self.angle,
                 (self.center.x, self.center.y))
        else:
            return "TextureSprite(format=%d, access=%d, size=%s, angle=%f)" % \
                (flags.value, access.value, (w.value, h.value), self.angle)



import gc
import sys
import pytest
from ctypes import addressof

from sdl2 import ext as sdl2ext
from sdl2 import dll
from sdl2.rect import SDL_Point
from sdl2.render import SDL_Renderer, SDL_Texture
from sdl2.surface import SDL_CreateRGBSurface, SDL_FreeSurface

# NOTE: These tests still have some legacy cruft to clean

def check_pixels(view, w, h, sprite, c1, c2, cx=0, cy=0):
    msg = "color mismatch at %d,%d: %d not in %s"
    cx = cx + sprite.x
    cy = cy + sprite.y
    cw, ch = sprite.size
    cmy = cy + ch
    cmx = cx + cw
    for y in range(w):
        for x in range(h):
            if cy <= y < cmy and cx <= x < cmx:
                assert view[y][x] == c1, msg % (x, y, view[y][x], c1)
            else:
                assert view[y][x] in c2, msg % (x, y, view[y][x], c2)

def check_areas(view, w, h, rects, c1, c2):
    def _inarea(x, y, rs):
        for r in rs:
            if (x >= r[0] and x < (r[0] + r[2]) and
                    y >= r[1] and y < (r[1] + r[3])):
                return True
        return False
    msg = "color mismatch at %d,%d: %d not in %s"
    for y in range(w):
        for x in range(h):
            if _inarea(x, y, rects):
                assert view[y][x] == c1, msg % (x, y, view[y][x], c1)
            else:
                assert view[y][x] in c2, msg % (x, y, view[y][x], c2)

def check_lines(view, w, h, points, c1, c2):
    def _online(x, y, pts):
        for p1, p2 in pts:
            if sdl2ext.point_on_line(p1, p2, (x, y)):
                return True
        return False
    msg = "color mismatch at %d,%d: %d not in %s"
    for y in range(w):
        for x in range(h):
            if _online(x, y, points):
                assert view[y][x] == c1, msg % (x, y, view[y][x], c1)
            else:
                assert view[y][x] in c2, msg % (x, y, view[y][x], c2)


class TestExtRenderer(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self, with_sdl):
        sf = SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)

        # Create renderer with SDL_Surface
        renderer = sdl2ext.Renderer(sf.contents)
        assert addressof(renderer.rendertarget) == addressof(sf.contents)
        assert isinstance(renderer.sdlrenderer.contents, SDL_Renderer)
        renderer.destroy()

        # Create renderer with SDL_Surface pointer
        renderer = sdl2ext.Renderer(sf)
        assert renderer.rendertarget == sf
        assert isinstance(renderer.sdlrenderer.contents, SDL_Renderer)
        renderer.destroy()

        # Create renderer with SoftwareSprite
        sprite = sdl2ext.SoftwareSprite(sf.contents, True)
        renderer = sdl2ext.Renderer(sprite)
        assert renderer.rendertarget == sprite
        assert isinstance(renderer.sdlrenderer.contents, SDL_Renderer)
        renderer.destroy()

        # Create renderer with Window
        window = sdl2ext.Window("Test", size=(1, 1))
        renderer = sdl2ext.Renderer(window)
        assert renderer.rendertarget == window
        assert isinstance(renderer.sdlrenderer.contents, SDL_Renderer)
        renderer.destroy()

        # Create renderer with SDL_Window
        sdlwindow = window.window
        renderer = sdl2ext.Renderer(sdlwindow)
        assert renderer.rendertarget == sdlwindow
        assert isinstance(renderer.sdlrenderer.contents, SDL_Renderer)
        renderer.destroy()
        del window

        # Test exception on using a destroyed renderer (and random type errors)
        with pytest.raises(RuntimeError):
            tst = renderer.sdlrenderer
        with pytest.raises(TypeError):
            sdl2ext.Renderer(None)
        with pytest.raises(TypeError):
            sdl2ext.Renderer(1234)
        with pytest.raises(TypeError):
            sdl2ext.Renderer("test")

    def test_logical_size(self, with_sdl):
        sf = SDL_CreateRGBSurface(0, 150, 50, 32, 0, 0, 0, 0)
        window = sdl2ext.Window("Test", size=(150, 50))
        sprite = sdl2ext.SoftwareSprite(sf.contents, True)

        targets = {
            "SDL_Surface": sf.contents,
            "SDL_Surface_pointer": sf,
            "SoftwareSprite": sprite,
            "Window": window,
            "SDL_Window": window.window
        }
        for name, target in targets.items():
            target_name = name
            renderer = sdl2ext.Renderer(target)
            assert isinstance(renderer.sdlrenderer.contents, SDL_Renderer)
            assert renderer.logical_size == (150, 50)
            renderer.logical_size = (200, 100)
            assert renderer.logical_size == (200, 100)
            renderer.reset_logical_size()
            assert renderer.logical_size == (150, 50)
            renderer.destroy()

        window.close()
        
    def test_color(self, with_sdl):
        sf = SDL_CreateRGBSurface(
            0, 10, 10, 32, 0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF
        )
        renderer = sdl2ext.Renderer(sf.contents)
        assert isinstance(renderer.color, sdl2ext.Color)
        assert renderer.color == sdl2ext.Color(0, 0, 0, 0)
        renderer.color = 0x00FF0000
        assert renderer.color == sdl2ext.Color(0xFF, 0, 0, 0)
        renderer.clear()
        view = sdl2ext.PixelView(sf.contents)
        check_areas(view, 10, 10, [[0, 0, 10, 10]], 0xFF000000, (0x0,))
        del view
        renderer.color = 0xAABBCCDD
        assert renderer.color == sdl2ext.Color(0xBB, 0xCC, 0xDD, 0xAA)
        renderer.clear()
        view = sdl2ext.PixelView(sf.contents)
        check_areas(view, 10, 10, [[0, 0, 10, 10]], 0xBBCCDDAA, (0x0,))
        del view
        renderer.destroy()
        SDL_FreeSurface(sf)

    @pytest.mark.skip("not implemented")
    def test_blendmode(self, with_sdl):
        pass

    @pytest.mark.skip("not implemented")
    def test_scale(self, with_sdl):
        pass

    def test_clear(self, with_sdl):
        sf = SDL_CreateRGBSurface(0, 10, 10, 32,
                                  0xFF000000,
                                  0x00FF0000,
                                  0x0000FF00,
                                  0x000000FF)
        renderer = sdl2ext.Renderer(sf.contents)
        assert isinstance(renderer.color, sdl2ext.Color)
        assert renderer.color == sdl2ext.Color(0, 0, 0, 0)
        renderer.color = 0x00FF0000
        assert renderer.color == sdl2ext.Color(0xFF, 0, 0, 0)
        renderer.clear()
        view = sdl2ext.PixelView(sf.contents)
        check_areas(view, 10, 10, [[0, 0, 10, 10]], 0xFF000000, (0x0,))
        del view
        renderer.clear(0xAABBCCDD)
        assert renderer.color == sdl2ext.Color(0xFF, 0, 0, 0)
        view = sdl2ext.PixelView(sf.contents)
        check_areas(view, 10, 10, [[0, 0, 10, 10]], 0xBBCCDDAA, (0x0,))
        del view
        renderer.destroy()
        SDL_FreeSurface(sf)

    def test_copy(self, with_sdl):
        # Initialize target surface and renderer
        surface = SDL_CreateRGBSurface(0, 128, 128, 32, 0, 0, 0, 0).contents
        renderer = sdl2ext.Renderer(surface)
        renderer.clear(0xAABBCC)
        view = sdl2ext.PixelView(surface)

        # Test copying a Texture without any arguments (should fill surface)
        sf = SDL_CreateRGBSurface(0, 16, 16, 32, 0, 0, 0, 0)
        sdl2ext.fill(sf, (0, 0, 0, 0))
        tx = sdl2ext.Texture(renderer, sf)
        renderer.copy(tx)
        renderer.present()
        assert view[0][0] == 0
        assert view[127][127] == 0

        # Test copying a Texture with only location argument
        renderer.clear(0xAABBCC) # reset surface
        renderer.copy(tx, dstrect=(10, 10))
        renderer.present()
        assert view[0][0] == 0xAABBCC
        assert view[10][10] == 0
        assert view[25][25] == 0
        assert view[26][26] == 0xAABBCC

        # Test copying a subset of a Texture
        renderer.clear(0xAABBCC) # reset surface
        renderer.copy(tx, srcrect=(0, 0, 10, 10), dstrect=(10, 10))
        renderer.present()
        assert view[0][0] == 0xAABBCC
        assert view[10][10] == 0
        assert view[19][19] == 0
        assert view[20][20] == 0xAABBCC

        # Test copying a Texture with location and size
        renderer.clear(0xAABBCC) # reset surface
        renderer.copy(tx, dstrect=(10, 10, 30, 40))
        renderer.present()
        assert view[0][0] == 0xAABBCC
        assert view[10][10] == 0
        assert view[49][39] == 0
        assert view[50][40] == 0xAABBCC

        if dll.version > 2005:
            # Test copying a Texture with rotation
            renderer.clear(0xAABBCC) # reset surface
            renderer.copy(tx, dstrect=(32, 32), angle=180, center=(0, 0))
            renderer.present()
            assert view[0][0] == 0xAABBCC
            assert view[16][16] == 0xFF000000  # Rotation suddenly adds alpha?
            assert view[31][31] == 0xFF000000  # Rotation suddenly adds alpha?
            assert view[32][32] == 0xAABBCC

        # (legacy): Test copying texture from a SpriteFactory
        renderer.clear(0)  # reset surface
        factory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        w, h = 32, 32
        sp = factory.from_color(0xFF0000, (w, h))
        sp.x, sp.y = 40, 50
        renderer.copy(sp, (0, 0, w, h), (sp.x, sp.y, w, h))
        check_pixels(view, 128, 128, sp, 0xFF0000, (0x0,))

        del view

    def test_draw_line(self, with_sdl):
        surface = SDL_CreateRGBSurface(0, 128, 128, 32, 0, 0, 0, 0).contents
        sdl2ext.fill(surface, 0x0)
        renderer = sdl2ext.Renderer(surface)
        renderer.draw_line((20, 10, 20, 86), 0x0000FF)
        view = sdl2ext.PixelView(surface)
        check_lines(view, 128, 128,
                         [((20, 10), (20, 86))], 0x0000FF, (0x0,))
        del view

    def test_draw_point(self, with_sdl):
        # Initialize target surface and renderer
        surface = SDL_CreateRGBSurface(0, 128, 128, 32, 0, 0, 0, 0).contents
        sdl2ext.fill(surface, 0x0)
        renderer = sdl2ext.Renderer(surface)

        # Try drawing a single point
        renderer.draw_point((1, 1), 0x0000FF)
        view = sdl2ext.PixelView(surface)
        assert view[1][1] == 0x0000FF

        # Try drawing a single SDL_Point
        renderer.draw_point(SDL_Point(3, 3), 0x0000FF)
        assert view[3][3] == 0x0000FF

        # Try drawing multiple points
        renderer.draw_point([(8, 8), SDL_Point(5, 5)], 0x0000FF)
        assert view[5][5] == 0x0000FF
        assert view[8][8] == 0x0000FF

        # Try subpixel rendering (if supported)
        if dll.version >= 2010:
            assert view[6][6] == 0x0
            renderer.draw_point((6.5, 6.5), 0x0000FF)
            assert view[6][6] != 0x0
        else:
            with pytest.raises(RuntimeError):
               renderer.draw_point((6.5, 6.5), 0x0000FF) 
        del view

        # Test exceptions on bad input
        with pytest.raises(ValueError):
            renderer.draw_point((0, 0, 2), 0x0000FF)
        with pytest.raises(ValueError):
            renderer.draw_point([(0, 0), (1, 2, 3)], 0x0000FF)
        
    def test_draw_rect(self, with_sdl):
        surface = SDL_CreateRGBSurface(0, 128, 128, 32, 0, 0, 0, 0).contents
        sdl2ext.fill(surface, 0x0)
        renderer = sdl2ext.Renderer(surface)
        renderer.draw_rect((40, 50, 32, 32), 0x0000FF)
        view = sdl2ext.PixelView(surface)
        check_lines(view, 128, 128, [
            ((40, 50), (71, 50)),
            ((40, 50), (40, 81)),
            ((40, 81), (71, 81)),
            ((71, 50), (71, 81))], 0x0000FF, (0x0,))
        del view
        sdl2ext.fill(surface, 0x0)
        renderer.draw_rect([(5, 5, 10, 10), (20, 15, 8, 10)], 0x0000FF)
        view = sdl2ext.PixelView(surface)
        check_lines(view, 128, 128, [
            ((5, 5), (14, 5)),
            ((5, 5), (5, 14)),
            ((5, 14), (14, 14)),
            ((14, 5), (14, 14)),
            ((20, 15), (27, 15)),
            ((20, 15), (20, 24)),
            ((20, 24), (27, 24)),
            ((27, 15), (27, 24))], 0x0000FF, (0x0,))
        del view

    def test_fill(self, with_sdl):
        surface = SDL_CreateRGBSurface(0, 128, 128, 32, 0, 0, 0, 0).contents
        sdl2ext.fill(surface, 0x0)
        renderer = sdl2ext.Renderer(surface)
        factory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        w, h = 32, 32
        sp = factory.from_color(0xFF0000, (w, h))
        sp.x, sp.y = 40, 50
        renderer.fill((sp.x, sp.y, w, h), 0x0000FF)
        view = sdl2ext.PixelView(surface)
        check_pixels(view, 128, 128, sp, 0x0000FF, (0x0,))
        del view

        sdl2ext.fill(surface, 0x0)
        renderer.fill([(5, 5, 10, 10), (20, 15, 8, 10)], 0x0000FF)
        view = sdl2ext.PixelView(surface)
        check_areas(view, 128, 128, [(5, 5, 10, 10), (20, 15, 8, 10)],
                         0x0000FF, (0x0,))
        del view


class TestExtTexture(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init_destroy(self):
        # Create renderer and test surface
        rendertarget = SDL_CreateRGBSurface(0, 100, 100, 32, 0, 0, 0, 0)
        renderer = sdl2ext.Renderer(rendertarget.contents)
        sf = SDL_CreateRGBSurface(
            0, 16, 16, 32, 0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF
        )

        # Test creation with surface pointer
        tx = sdl2ext.Texture(renderer, sf)
        assert isinstance(tx.tx.contents, SDL_Texture)

        # Test destruction and associated behaviour
        tx.destroy()
        with pytest.raises(RuntimeError):
            sdl_tx = tx.tx

        # Test creation with surface contents
        tx = sdl2ext.Texture(renderer, sf.contents)
        assert isinstance(tx.tx.contents, SDL_Texture)
        
        # Test texture size
        assert tx.size == (16, 16)

        # Test exception on bad input
        with pytest.raises(TypeError):
            sdl2ext.Texture(sf, sf)
        with pytest.raises(TypeError):
            sdl2ext.Texture(renderer, renderer)

        # Test exception when accessing Texture with destroyed Renderer
        renderer.destroy()
        with pytest.raises(RuntimeError):
            sdl_tx = tx.tx
        tx.destroy()  # Ensure texture destruction method doesn't error
        SDL_FreeSurface(sf)

    @pytest.mark.skip("not implemented")
    def test_size(self, with_sdl):
        pass

    @pytest.mark.skip("not implemented")
    def test_scale_mode(self, with_sdl):
        pass


@pytest.mark.skip("not implemented")
def test_set_texture_scale_quality(with_sdl):
    pass



import sys
import pytest
import sdl2
from ctypes import create_string_buffer, byref
from sdl2 import error

# Override global cleanup fixture since it calls SDL_ClearError
@pytest.fixture(autouse=True)
def sdl_cleanup():
    yield


def test_SDL_GetSetClearError():
    error.SDL_SetError(b"Hi there!")
    assert error.SDL_GetError() == b"Hi there!"
    error.SDL_SetError(b"Error 2");
    assert error.SDL_GetError() == b"Error 2"
    error.SDL_ClearError();
    assert error.SDL_GetError() == b""

@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_GetErrorMsg():
    error.SDL_SetError(b"123456789")
    assert error.SDL_GetError() == b"123456789"
    buf = create_string_buffer(10)
    assert error.SDL_GetErrorMsg(buf, 10) == b"123456789"
    assert buf.value == b"123456789"
    buf2 = create_string_buffer(5)
    assert error.SDL_GetErrorMsg(buf2, 5) == b"1234"
    assert buf2.value == b"1234"
    error.SDL_ClearError()



from .compat import isiterable
from .common import raise_sdl_err
from .color import convert_to_color
from .. import pixels
from .. import surface as surf
from ..rect import SDL_Rect

__all__ = ["subsurface"]


def _get_rect_tuple(r, argname):
    if isinstance(r, SDL_Rect):
        return (r.x, r.y, r.w, r.h)
    elif isiterable(r) and len(r) == 4:
        return tuple(r)
    else:
        e = "'{0}' must be an SDL_Rect or tuple of 4 integers."
        raise TypeError(e.format(argname))


def _get_target_surface(target, argname="target"):
    """Gets the SDL_surface from the passed target."""
    if hasattr(target, "surface"):  # i.e. if SoftwareSprite
        rtarget = target.surface  
    elif isinstance(target, surf.SDL_Surface):
        rtarget = target
    elif "SDL_Surface" in str(type(target)):
        rtarget = target.contents
    else:
        raise TypeError(
            "{0} must be a valid Sprite or SDL Surface".format(argname)
        )
    return rtarget


def _create_surface(size, fill=None, fmt="ARGB8888", errname="SDL"):
    # Perform initial type and argument checking
    if not isiterable(size) and len(size) == 2:
        e = "Surface size must be a tuple of two positive integers."
        raise TypeError(e)
    if not all([i > 0 and int(i) == i for i in size]):
        e = "Surface height and width must both be positive integers (got {0})."
        raise ValueError(e.format(str(size)))
    if fmt not in pixels.NAME_MAP.keys() and fmt not in pixels.ALL_PIXELFORMATS:
        e = "'{0}' is not a supported SDL pixel format."
        raise ValueError(e.format(fmt))
    if fill is not None:
        fill = convert_to_color(fill)

    # Actually create a surface with the given pixel format
    w, h = size
    bpp = 32  # NOTE: according to the SDL_surface.c code, this has no effect
    fmt_enum = fmt if type(fmt) == int else pixels.NAME_MAP[fmt]
    sf = surf.SDL_CreateRGBSurfaceWithFormat(0, w, h, bpp, fmt_enum)
    if not sf:
        raise_sdl_err("creating the {0} surface".format(errname))

    # If provided, set the fill for the new surface
    if fill is not None:
        pixfmt = sf.contents.format.contents
        if pixfmt.Amask == 0:
            fill_col = pixels.SDL_MapRGB(pixfmt, fill.r, fill.g, fill.b)
        else:
            fill_col = pixels.SDL_MapRGBA(pixfmt, fill.r, fill.g, fill.b, fill.a)
        surf.SDL_FillRect(sf, None, fill_col)

    return sf


def subsurface(surface, area):
    """Creates a new :obj:`~sdl2.SDL_Surface` from a part of another surface.

    Surfaces created with this function will share pixel data with the original
    surface, meaning that any modifications to one surface will result in
    modifications to the other.

    .. warning::
       Because subsurfaces share pixel data with their parent surface, they
       *cannot* be used after the parent surface is freed. Doing so will
       almost certainly result in a segfault.

    Args:
        surface (:obj:`~sdl2.SDL_Surface`): The parent surface from which
            new sub-surface should be created.
        area (:obj:`SDL_Rect`, tuple): The ``(x, y, w, h)`` subset of the parent
            surface to use for the new surface, where ``x, y`` are the pixel
            coordinates of the top-left corner of the rectangle and ``w, h`` are
            its width and height (in pixels). Can also be specified as an
            :obj:`SDL_Rect`.

    Returns:
        :obj:`~sdl2.SDL_Surface`: The newly-created subsurface.

    """
    if not isinstance(surface, surf.SDL_Surface):
        if "SDL_Surface" in str(type(surface)):
            surface = surface.contents
        else:
            e = "'surface' must be an SDL_Surface (got {0})"
            raise TypeError(e.format(type(surface)))

    x, y, w, h = _get_rect_tuple(area, argname="area")
    if x + w > surface.w or y + h > surface.h:
        e = "The specified area {0} exceeds the bounds of the parent surface "
        e += str((surface.w, surface.h))
        raise ValueError(e.format(str(area)))

    fmt = surface.format[0]
    bpp = fmt.BitsPerPixel
    subpixels = (surface.pixels + surface.pitch * y + fmt.BytesPerPixel * x)
    subsurf = surf.SDL_CreateRGBSurfaceFrom(
        subpixels, w, h, bpp, surface.pitch, fmt.Rmask, fmt.Gmask, fmt.Bmask, fmt.Amask
    )
    if not subsurf:
        raise_sdl_err("creating the subsurface")
        
    return subsurf.contents



from ctypes import POINTER, c_char_p, c_int, c_float, c_void_p
from .dll import _bind
from .stdinc import Uint32


__all__ = [
    # Structs
    "SDL_Sensor",

    # Defines
    "SDL_SensorID", "SDL_STANDARD_GRAVITY",

    # Enums
    "SDL_SensorType",
    "SDL_SENSOR_INVALID", "SDL_SENSOR_UNKNOWN", "SDL_SENSOR_ACCEL",
    "SDL_SENSOR_GYRO",

    # Functions
    "SDL_LockSensors", "SDL_UnlockSensors", "SDL_NumSensors",
    "SDL_SensorGetDeviceName", "SDL_SensorGetDeviceType",
    "SDL_SensorGetDeviceNonPortableType", "SDL_SensorGetDeviceInstanceID",
    "SDL_SensorOpen", "SDL_SensorFromInstanceID", "SDL_SensorGetName",
    "SDL_SensorGetType", "SDL_SensorGetNonPortableType",
    "SDL_SensorGetInstanceID", "SDL_SensorGetData", "SDL_SensorClose",
    "SDL_SensorUpdate"
]


class SDL_Sensor(c_void_p):
    pass

SDL_SensorID = Uint32

SDL_SensorType = c_int
SDL_SENSOR_INVALID = -1
SDL_SENSOR_UNKNOWN = 0
SDL_SENSOR_ACCEL = 1
SDL_SENSOR_GYRO = 2

SDL_STANDARD_GRAVITY = 9.80665

SDL_LockSensors = _bind("SDL_LockSensors", None, None, added='2.0.14')
SDL_UnlockSensors = _bind("SDL_UnlockSensors", None, None, added='2.0.14')
SDL_NumSensors = _bind("SDL_NumSensors", None, c_int, added='2.0.9')
SDL_SensorGetDeviceName = _bind("SDL_SensorGetDeviceName", [c_int], c_char_p, added='2.0.9')
SDL_SensorGetDeviceType = _bind("SDL_SensorGetDeviceType", [c_int], SDL_SensorType, added='2.0.9')
SDL_SensorGetDeviceNonPortableType = _bind("SDL_SensorGetDeviceNonPortableType", [c_int], c_int, added='2.0.9')
SDL_SensorGetDeviceInstanceID = _bind("SDL_SensorGetDeviceInstanceID", [c_int], SDL_SensorID, added='2.0.9')
SDL_SensorOpen = _bind("SDL_SensorOpen", [c_int], POINTER(SDL_Sensor), added='2.0.9')
SDL_SensorFromInstanceID = _bind("SDL_SensorFromInstanceID", [SDL_SensorID], POINTER(SDL_Sensor), added='2.0.9')
SDL_SensorGetName = _bind("SDL_SensorGetName", [POINTER(SDL_Sensor)], c_char_p, added='2.0.9')
SDL_SensorGetType = _bind("SDL_SensorGetType", [POINTER(SDL_Sensor)], SDL_SensorType, added='2.0.9')
SDL_SensorGetNonPortableType = _bind("SDL_SensorGetNonPortableType", [POINTER(SDL_Sensor)], c_int, added='2.0.9')
SDL_SensorGetInstanceID = _bind("SDL_SensorGetInstanceID", [POINTER(SDL_Sensor)], SDL_SensorID, added='2.0.9')
SDL_SensorGetData = _bind("SDL_SensorGetData", [POINTER(SDL_Sensor), POINTER(c_float), c_int], c_int, added='2.0.9') # Needs testing
SDL_SensorClose = _bind("SDL_SensorClose", [POINTER(SDL_Sensor)], None, added='2.0.9')
SDL_SensorUpdate = _bind("SDL_SensorUpdate", None, None, added='2.0.9')
